// Pick a point in the parents then grab alternating chunks for each child.
//   A word about crossover site mapping.  If a genome has width 10, the
// cross site can assume a value of 0 to 10, inclusive.  A site of 0 means
// that all of the material comes from the father.  A site of 10 means that
// all of the material comes from the mother.  A site of 3 means that bits
// 0-2 come from the mother and bits 3-9 come from the father.

// smd 9/7/04
// modified to only produce crossover at points marking 
// codon boundaries
// codon size is kept in CODONSIZE static variable

// New GE crossover
// uses Lance's idea to preserve building blocks 
// by crossing over between same nonterminals
// and then only crossing with the functional subunit
// for that nonterminal

// step 1:  determine which chromosome is starter
// step 2:  select crossover point
// step 3:  randomly select crossover point on second chromosome
// step 4:  if not same nonterminal:  randomly search up or downstream
//          to find an equivalent point
// step 5:  if still no match search opposite direction from starting point
// step 6:  if no match, end with no crossover
// step 7:  if match determine end of functional block for both chromsomes
//          and swap 

int
GEBinStrBlockCrossGenome::
BlockPreserveCrossover(const GAGenome& p1, const GAGenome& p2,
		  GAGenome* c1, GAGenome* c2){
  const GEBinStrBlockCrossGenome &mom=
    DYN_CAST(const GEBinStrBlockCrossGenome &, p1);
  const GEBinStrBlockCrossGenome &dad=
    DYN_CAST(const GEBinStrBlockCrossGenome &, p2);

  int n=0;
  unsigned int momsite, momlen;
  unsigned int dadsite, dadlen;

  if(c1 && c2){
    GEBinStrBlockCrossGenome &sis=DYN_CAST(GEBinStrBlockCrossGenome &, *c1);
    GEBinStrBlockCrossGenome &bro=DYN_CAST(GEBinStrBlockCrossGenome &, *c2);

    if(sis.resizeBehaviour() == GAGenome::FIXED_SIZE &&
       bro.resizeBehaviour() == GAGenome::FIXED_SIZE){
      if(mom.length() != dad.length() || 
	 sis.length() != bro.length() ||
	 sis.length() != mom.length()){
	GAErr(GA_LOC, mom.className(), "one-point cross", gaErrSameLengthReqd);
	return n;
      }
      momsite = dadsite = GEBinStrBlockCrossGenome::codon_to_orig(GARandomInt(0, GEBinStrBlockCrossGenome::orig_to_codon(mom.length())));
      momlen = dadlen = mom.length() - momsite;
    }
    else if(sis.resizeBehaviour() == GAGenome::FIXED_SIZE ||
	    bro.resizeBehaviour() == GAGenome::FIXED_SIZE){
      GAErr(GA_LOC, mom.className(), "one-point cross", gaErrSameBehavReqd);
      return n;
    }
    else{
      momsite = GEBinStrBlockCrossGenome::codon_to_orig(GARandomInt(0, GEBinStrBlockCrossGenome::orig_to_codon(mom.length())));
      dadsite = GEBinStrBlockCrossGenome::codon_to_orig(GARandomInt(0, GEBinStrBlockCrossGenome::orig_to_codon(dad.length())));
      momlen = mom.length() - momsite;
      dadlen = dad.length() - dadsite;
      sis.resize(momsite+dadlen);
      bro.resize(dadsite+momlen);
    }

    sis.copy(mom, 0, 0, momsite);
    sis.copy(dad, momsite, dadsite, dadlen);
    bro.copy(dad, 0, 0, dadsite);
    bro.copy(mom, dadsite, momsite, momlen);
  
    n = 2;
  }
  else if(c1 || c2){
    GEBinStrBlockCrossGenome &sis = (c1 ? 
				   DYN_CAST(GEBinStrBlockCrossGenome&, *c1) : 
				   DYN_CAST(GEBinStrBlockCrossGenome&, *c2));

    if(sis.resizeBehaviour() == GAGenome::FIXED_SIZE){
      if(mom.length() != dad.length() || sis.length() != mom.length()){
	GAErr(GA_LOC, mom.className(), "one-point cross", gaErrSameLengthReqd);
	return n;
      }
      momsite = dadsite = GEBinStrBlockCrossGenome::codon_to_orig(GARandomInt(0, GEBinStrBlockCrossGenome::orig_to_codon(mom.length())));
      momlen = dadlen = mom.length() - momsite;
    }
    else{
      momsite = GEBinStrBlockCrossGenome::codon_to_orig(GARandomInt(0, GEBinStrBlockCrossGenome::orig_to_codon(mom.length())));
      dadsite = GEBinStrBlockCrossGenome::codon_to_orig(GARandomInt(0, GEBinStrBlockCrossGenome::orig_to_codon(dad.length())));
      momlen = mom.length() - momsite;
      dadlen = dad.length() - dadsite;
      sis.resize(momsite+dadlen);
    }

    if(GARandomBit()){
      sis.copy(mom, 0, 0, momsite); // copies all the way to end from mother
      sis.copy(dad, momsite, dadsite, dadlen);
    }
    else{
      sis.copy(dad, 0, 0, dadsite);
      sis.copy(mom, dadsite, momsite, momlen);
    }

    n = 1;
  }

  return n;
}