/// Config.cpp

#include "Config.h"

///
/// Default constructor 
///
Config::Config(){
  init_seed = (unsigned int)(time(NULL));
  init();
}

///
/// Alternative constructor that reads in the configuration file
///
Config::Config(string pref_file){
  init_seed = (unsigned int)(time(NULL));
  init();
  read_config(pref_file);
}

///
/// Initialize maps for configuration
///
void Config::init(){
  select_type = "uniform";
  SelectionTypeMap["uniform"] = uniformSelection;
  SelectionTypeMap["tournament"] = tournamentSelection;
  SelectionTypeMap["roulette"] = rouletteSelection;
  SelectionTypeMap["rank"] = rankSelection;
  cross_type = "standard";
  CrossoverTypeMap["standard"] = standardCross;
  CrossoverTypeMap["matchCross"] = matchCross;
  CrossoverTypeMap["matchblockCross"] = matchblockCross;
  algorithm_type = "simple";
  AlgorithmTypeMap["simple"] = simpleAlgorithm;
  AlgorithmTypeMap["random"] = randomAlgorithm;
  fitness_type = "balanced";
  FitnessTypeMap["classification"] = classError;
  FitnessTypeMap["balanced"] = balancedAcc;
	FitnessTypeMap["weighted"] = weightedCE;
	FitnessTypeMap["rsquared"] = rSquared;

	/*mmouble means all input values are stored as doubles,
		mmgenotypeMiser means they are ott-dummy values
		which are packed two to a byte*/
	memory_manager = "mmdouble";
	MemoryManagerMap["mmdouble"] = mmdouble;
	MemoryManagerMap["mmgenotypeMiser"] = mmgenotypeMiser;
}


///
/// sets a map containing the parameters read in from the config
/// file
/// @param configname string with filename of configuration file
///
void Config::read_config(string configname = ""){
  map<string, string> pref_map;

  //set defaults
  pref_map[SEED] = Stringmanip::itos(init_seed);
  pref_map[MAXGEN] = Stringmanip::itos(DEFAULT_GENS);
  pref_map[POPSIZE] = Stringmanip::itos(DEFAULT_POP);
  pref_map[DEMES] = Stringmanip::itos(DEFAULT_DEMES);
  pref_map[MUTATE] = Stringmanip::itos(DEFAULT_MUTATION);
  pref_map[CROSS] = Stringmanip::itos(DEFAULT_CROSSOVER);
  pref_map[UPDATE] = Stringmanip::itos(DEFAULT_UPDATE);
  pref_map[CODONSIZE] = Stringmanip::itos(DEFAULT_CODONSIZE);
  pref_map[WRAPCOUNT] = Stringmanip::itos(DEFAULT_WRAPCOUNT);
  pref_map[MINCHROM] = Stringmanip::itos(DEFAULT_MINCHROM);
  pref_map[MAXCHROM] = Stringmanip::itos(DEFAULT_MAXCHROM);
  pref_map[PVMEXCHANGEGEN] = Stringmanip::itos(DEFAULT_PVMEXCHANGEGEN);
  pref_map[INITDEPTH] =Stringmanip::itos(DEFAULT_INITDEPTH);
  pref_map[SELECTIONTYPE] = select_type;
  pref_map[CROSSOVERTYPE] = cross_type;
  pref_map[ALGORITHMTYPE] = algorithm_type;
  pref_map[DEBUG] = Stringmanip::itos(DEFAULT_DEBUG);
  pref_map[FITNESSTYPE] = fitness_type;
  pref_map[MISSINGVALUE] = Stringmanip::itos(DEFAULT_MISSING);
	pref_map[MEMORYMANAGER] = memory_manager;
	
  
  ifstream config(configname.c_str(), ios::in);
  if(!config.is_open()){
    throw ErrorExcept(configname + ":  cannot open!");
  }

  string key, value;

  // read in each parameter to use as key in map and its parameter
  string blank;  // use to skip = sign in file
  while(!config.eof()){
    config >> key;
    // check that line doesn't begin with a # 
    if(key[0] != '#'){
      config >> blank >> value;
      pref_map[key] = value; // insert pair into map
    }
    skip_rest(config); // skip any comments after values
    key="";
  }
  config.close();

  // now set the parameters using the map
  set_params(pref_map);
}

///
/// sets variables within object to match parameters
/// @param pref_map map having keys as configuration keys
/// and values from the configuration file
/// @throws LapExcept if no parameter found
///
void Config::set_params(map<string, string> & pref_map){
        // construct a string to use in assigning values
        string par_holder;

   par_holder += get_param(SEED, pref_map) + ' ';
   par_holder += get_param(MAXGEN, pref_map) + ' ';
   par_holder += get_param(POPSIZE, pref_map) + ' ';
   par_holder += get_param(DEMES, pref_map) + ' ';
   par_holder += get_param(MUTATE, pref_map) + ' ';
   par_holder += get_param(CROSS, pref_map) + ' ';
   par_holder += get_param(UPDATE, pref_map) + ' ';
   par_holder += get_param(CODONSIZE, pref_map) + ' ';
   par_holder += get_param(WRAPCOUNT, pref_map) + ' ';
   par_holder += get_param(MINCHROM, pref_map) + ' ';
   par_holder += get_param(MAXCHROM, pref_map) + ' ';
   par_holder += get_param(TRAINFILE, pref_map) + ' ';
   par_holder += get_param(TESTFILE, pref_map) + ' ';
   par_holder += get_param(GRAMFILE, pref_map) + ' ';
   par_holder += get_param(BASEOUT, pref_map) + ' ';
   par_holder += get_param(PVMEXCHANGEGEN, pref_map) + ' ';
   par_holder += get_param(INITDEPTH, pref_map) + ' ';
   par_holder += get_param(SELECTIONTYPE, pref_map) + ' ';
   par_holder += get_param(CROSSOVERTYPE, pref_map) + ' ';
   par_holder += get_param(ALGORITHMTYPE, pref_map) + ' ';
   par_holder += get_param(DEBUG, pref_map) + ' ';
   par_holder += get_param(FITNESSTYPE, pref_map) + ' ';
   par_holder += get_param(MISSINGVALUE, pref_map) + ' ';
	 par_holder += get_param(MEMORYMANAGER, pref_map) + ' ';

  stringstream ss(par_holder);
  ss >> rand_seed >> num_gens >> pop_size >> num_demes >> mutation_rate >> 
    crossover_rate >> update_interval >> codon_size >> wrapper_count >> min_chrom >>
    max_chrom >> trainfile >> testfile >> gramfile >> base_output >> pvm_exchange_interval
		 >> initdepth >> select_type >> cross_type >> algorithm_type >> debugstatus >> fitness_type >> missing_value >> memory_manager;

  check_params();
}


///
/// Checks parameters are valid
/// @throws ErrorExcept if any error detected
///
void  Config::check_params(){
  if(rand_seed > SEED_MAX || rand_seed < 0)
    throw ErrorExcept(SEED + " must be greater than 0 and less than " + Stringmanip::itos(SEED_MAX));
  if(crossover_rate <= 0 || crossover_rate > 1)
   throw ErrorExcept(CROSS + " must be greater than 0 and less than 1");
  if(mutation_rate <= 0 || mutation_rate > 1)
   throw ErrorExcept(MUTATE + " must be greater than 0 and less than 1");
  if(num_demes <= 0)
   throw ErrorExcept(DEMES + " must be greater than 0");
  if(num_gens <= 0)
   throw ErrorExcept(MAXGEN + " must be greater than 0");
  if(pop_size <= 0)
   throw ErrorExcept(POPSIZE + " must be greater than 0");
  if(update_interval <= 0)
   throw ErrorExcept(UPDATE + " must be greater than 0");
  if(wrapper_count <= 0)
   throw ErrorExcept(WRAPCOUNT + " must be greater than 0");
  if(min_chrom <= 0)
   throw ErrorExcept(MINCHROM + " must be greater than 0");
  if(max_chrom < min_chrom)
   throw ErrorExcept(MAXCHROM + " must be greater than or equal to " + MINCHROM);
  if(codon_size < 1)
   throw ErrorExcept(CODONSIZE + " must be greater than 0");  
  if(pvm_exchange_interval <= 0)
   throw ErrorExcept(PVMEXCHANGEGEN + " must be greater than 0");
  if(initdepth < 0)
   throw ErrorExcept(INITDEPTH + " must be greater than or equal to 0");
  if(!valid_selection(select_type)){
    throw ErrorExcept(SELECTIONTYPE + " must be uniform, roulette, tournament, or rank");
  }
  if(!valid_crossover(cross_type)){
    throw ErrorExcept(CROSSOVERTYPE + " must be standard, match, or matchblock not " + cross_type);
  }
  if(!valid_algorithm(algorithm_type)){
    throw ErrorExcept(ALGORITHMTYPE + " must be simple or random not " + algorithm_type);
  }
	if(!valid_memory(memory_manager)) {
		throw ErrorExcept(MEMORYMANAGER + " must be mmdouble or mmgenotypeMiser not " + memory_manager);
	}
  if(debugstatus != 0 && debugstatus != 1){
    throw ErrorExcept(DEBUG + " must be either 1 (on) or 0 (off) not " + Stringmanip::itos(debugstatus));
  }
  if(!valid_fitness(fitness_type)){
    throw ErrorExcept(FITNESSTYPE + " must be simple or random not " + FITNESSTYPE);
  }  
  if(missing_value >= -1 && missing_value <= 2){
    throw ErrorExcept(MISSINGVALUE + " must be less than -1 or greater than 2");
  }
}


///
/// Checks that the selection type is valid
/// @param selectType string containing fitness type 
/// @return true if valid and false if not
///
bool Config::valid_selection(string selectType){
  if(SelectionTypeMap.find(selectType) != SelectionTypeMap.end()){
    return true;
  }
  return false;
}


///
/// Checks that the crossover type is valid   
/// @param crossType string containing fitness type 
/// @return true if valid and false if not
///
bool Config::valid_crossover(string crossType){
  if(CrossoverTypeMap.find(crossType) != CrossoverTypeMap.end()){
    // reset wrapper count to 1 when not standard crossover used
    if(CrossoverTypeMap[crossType] != standardCross)
      wrapper_count=1;
    return true;
  }
  return false;
}


///
/// Checks that the alorithm is valid
/// @param selectType string containing fitness type
/// @return true if valid and false if not
///
bool Config::valid_algorithm(string selectType){
  if(AlgorithmTypeMap.find(selectType) != AlgorithmTypeMap.end()){
    return true;
  }
  return false;
}


///
/// Checks that the fitnessType is valid
/// @param fitnessType string containing fitness type
/// @return true if valid and false if not
///
bool Config::valid_fitness(string fitnessType){
  if(FitnessTypeMap.find(fitnessType) != FitnessTypeMap.end()){
    return true;
  }
  return false;
}

bool Config::valid_memory(string memory_manager) {
	if (MemoryManagerMap.find(memory_manager) != MemoryManagerMap.end()) {
		return true;
	}
	return false;
}

///
/// returns string with map element or throws ErrorExcept if doesn't
/// exist
/// @param key string  
/// @param params map containing keys and values from configuration
/// file
/// @return value for the key passed 
///
string Config::get_param(string key, map<string, string> & params){
  map<string, string>::iterator i = params.find(key);
  if(i == params.end()){
    string err = "Missing parameter " + key;
    throw ErrorExcept(err);
  }
  return i->second;  // returning string that is held in the map for the key
}

///
/// skips rest of a line of input
/// @param  filestream stream to skip line
///
void Config::skip_rest(ifstream & filestream){
        char c;
        do{
                c = filestream.get();
        }while(c != '\n' && c != EOF);
}

///
///  Following are all accessor functions
///
unsigned int Config::get_gen() const{
   return num_gens;
}

unsigned int Config::get_demes() const{
   return num_demes;
}

unsigned int Config::get_pop_size() const{
   return pop_size;
}

double Config::get_mutate() const{
   return mutation_rate;
}

double Config::get_cross() const{
   return crossover_rate;
}

int Config::get_update_interval() const{
   return update_interval;
}


int Config::get_codon_size() const{
  return codon_size;
}

int Config::get_wrapper_count() const{
  return wrapper_count;
}

int Config::get_min_chrom() const{
  return min_chrom;
}

int Config::get_max_chrom() const{
  return max_chrom;
}
    
string Config::get_testfile() const{
  return testfile;
}

string Config::get_trainfile() const{
  return trainfile;
}

string Config::get_gramfile() const{
  return gramfile;
}

string Config::get_base_output() const{
  return base_output;
}

int Config::get_pvm_update_interval() const{
  return pvm_exchange_interval;
}

int Config::get_init_depth() const{
  return initdepth;
}

SelectionTypes Config::get_selection_type(){
  return SelectionTypeMap[select_type];
}

CrossoverTypes Config::get_crossover_type(){
  return CrossoverTypeMap[cross_type];
}

AlgorithmTypes Config::get_algorithm_type(){
  return AlgorithmTypeMap[algorithm_type];
}

FitnessTypes Config::get_fitness_type(){
  return FitnessTypeMap[fitness_type];
}

MemoryManagers Config::get_memory_manager() {
	return MemoryManagerMap[memory_manager];
}

int Config::get_debug() const{
  return debugstatus;
}

int Config::get_missing_value() const{
  return missing_value;
}

///
/// Returns seed number 
/// @returns see number 
///
unsigned int Config::get_seed() const{
  return rand_seed;
}


