// Config.h


#ifndef __CONFIG_H__
#define __CONFIG_H__

#include <iostream>
#include <fstream>
#include <sstream>
#include <map>
#include <vector>
#include <limits.h>
#include "Stringmanip.h"

using namespace std;

#include "ErrorExcept.h"

const string CROSS= "crossover_rate", MUTATE = "mutation_rate", SEED = "random_seed",
  MAXGEN = "max_generations", POPSIZE = "pop_size", CODONSIZE = "codon_size", WRAPCOUNT = "wrapper_count",
  MINCHROM = "min_chrom_size", MAXCHROM = "max_chrom_size", UPDATE = "update_interval",
  TRAINFILE = "train_file", TESTFILE = "test_file", GRAMFILE = "grammar_file", DEMES = "num_demes",
  BASEOUT = "output.basename", PVMEXCHANGEGEN = "pvm_exchange_generation", INITDEPTH = "init_depth",
  SELECTIONTYPE = "selection", CROSSOVERTYPE = "crossover", ALGORITHMTYPE = "algorithm",
  DEBUG = "debug", FITNESSTYPE = "fitness", MISSINGVALUE = "missing_value", MEMORYMANAGER="memory_manager";

/// specifies fitness types available
enum FitnessTypes{
  classError,
  balancedAcc,
	weightedCE,
	rSquared
};

/// Selection types available for evolution
enum SelectionTypes{ 
  uniformSelection,
  tournamentSelection,
  rouletteSelection,
  rankSelection,
};
    
/// Crossover available 
enum CrossoverTypes{
  standardCross,
  matchCross,
  matchblockCross,
};
    
/// Algorithms that can be run 
enum AlgorithmTypes{
  simpleAlgorithm,
  randomAlgorithm
};

/// Input variable handling (memory management).
/// mmError is a default value that should _never_ be actually selected
enum MemoryManagers {
	mmdouble,
	mmgenotypeMiser,
	mmError
};

///
/// Reads Config file for setting configuration parameters in GENN
///
    
class Config{

  public:
    Config();
    Config(string config_file);

    /// Use to read in configuration file and set the parameters
    void read_config(string configname); 

    // Accessor functions for various parameters
    unsigned int get_seed() const;
    unsigned int get_gen() const;
    unsigned int get_demes() const;
    unsigned int get_pop_size() const;
    int get_codon_size() const;
    int get_wrapper_count() const;
    int get_min_chrom() const;
    int get_max_chrom() const;
    int get_debug() const;
    int get_missing_value() const;
    string get_testfile() const;
    string get_trainfile() const;
    string get_gramfile() const;
    double get_mutate() const;
    double get_cross() const;
    int get_update_interval() const;
    int get_pvm_update_interval() const;
    int get_init_depth() const;
    string get_base_output() const;
    SelectionTypes get_selection_type();
    CrossoverTypes get_crossover_type();
    AlgorithmTypes get_algorithm_type();
    FitnessTypes get_fitness_type();
		MemoryManagers get_memory_manager();
    
  private:
    /// assigns values to parameters based on reading of configuration file
    void set_params();

    /// skips remainder of a line after parameters are read
    void skip_rest(ifstream & filestream);

    /// returns the value as a string of the parameter in the log file
    string get_param(string key, map<string, string> & params);

    /// sets the variables to be equal to values in config file
    void set_params(map<string, string> & pref_map);

    /// checks that parameters are set and within range
    void check_params();

    /// checks that a valid selector has been used in config file
    bool valid_selection(string selectType);

    /// checks that a valid crossover type has been used in config file
    bool valid_crossover(string crossType);

    /// checks that a valid algorithm has been selected
    bool valid_algorithm(string algType);

    /// checks that a valid fitness has been specified
    bool valid_fitness(string fitnessType);

		//checks that a valid memory manager has been specified
		bool valid_memory(string memoryManager);

    /// performs initialization
    void init();

    unsigned int rand_seed, init_seed,num_gens, num_demes, pop_size;
    double mutation_rate, crossover_rate;
    int update_interval, codon_size, wrapper_count, min_chrom, max_chrom, 
      pvm_exchange_interval, initdepth, debugstatus, missing_value;
    string testfile, trainfile, gramfile, base_output, select_type, cross_type,
      algorithm_type, fitness_type, memory_manager;
    string config_file;
    map<string, SelectionTypes> SelectionTypeMap;
    map<string, CrossoverTypes> CrossoverTypeMap;
    map<string, AlgorithmTypes> AlgorithmTypeMap;
    map<string, FitnessTypes> FitnessTypeMap;
		map<string, MemoryManagers> MemoryManagerMap;
    
    const static unsigned int SEED_MAX = INT_MAX;
    int static const DEFAULT_POP = 100;
    int static const DEFAULT_DEMES = 10;
    int static const DEFAULT_GENS = 50;
    int static const DEFAULT_UPDATE = 10;
    int static const DEFAULT_MINCHROM = 50;
    int static const DEFAULT_MAXCHROM = 100;
    int static const DEFAULT_WRAPCOUNT = 1;
    int static const DEFAULT_CODONSIZE = 8;
    int static const DEFAULT_PVMEXCHANGEGEN = 25;
    int static const DEFAULT_INITDEPTH = 0;  // when zero no sensible initialization
    int static const DEFAULT_DEBUG = 0;
    int static const DEFAULT_MISSING = 3;

    double static const DEFAULT_MUTATION = 0.01;
    double static const DEFAULT_CROSSOVER = 0.9;
};

#endif
