//Grammar.cpp

#include "Grammar.h"

///
/// Constructor creates basic syntax elements
///
Grammar::Grammar(){
  // create comma and parentheses as terminals
  element_map["("] = new Element("(", true);
  element_map[")"] = new Element(")", true);
  element_map[","] = new Element(",", true);
  initTreeDepth=0;
}

///
/// Destructor frees element memory
///
Grammar::~Grammar(){
  map<string, Element* >::iterator i;
  for(i=element_map.begin(); i!=element_map.end(); i++)
    delete i->second;
}

/// 
/// Parses grammar file and creates elements
/// @param filename string for grammar file
///
void Grammar::read_gram_file(string filename){

  ifstream gramfile(filename.c_str(), ios::in);
  if(!gramfile.is_open()){
    throw ErrorExcept(filename + ":  cannot open!");
  }

  bool start_found = false, terminals_defined = false, nonterminals_defined = false;
  string rule_name;
  
  // contains all lines related to grammar
  // all comments are skipped 
  // and the syntax is cleaned so that spaces are between all elements
  vector<string> useful_lines; 

  char line[LINEMAX];

  while(!gramfile.eof()){
    gramfile.getline(line, LINEMAX);

    string current_line(line);

    int linelength;
    if(!(linelength = current_line.length())){
      continue; // blank line so go to next one in file
    }
    
    //skip any whitespace
    unsigned int i;
    for(i=0; i<current_line.length(); i++){
      if(current_line[i] != ' ' && current_line[i] != '\t')
        break;
    }

    // check for comments (should begin with '//' or #
    // skip to next otherwise
    if(current_line[i] == '#' || (current_line[i] == '/' && current_line[i+1] == '/')) {
      continue;    
    }

    //ensure that all symbols are separated by at least one space
    insert_spaces(current_line);
   
    vector<string> tokens;
    //split the string on either space or tab
    if(split_string(current_line, " ", "\t", tokens) > 0){
      // N = for nonterminal list
      // T = for terminal list
      // S = for start expression
      // these must appear before any of the production rules
      //cerr << "Grammar::read_gram_file line: " << current_line << "\n";
      if(tokens[0] == "N"){
        set_nonterminals(tokens);
        nonterminals_defined = true;
	//cerr << "Nonterminals defined\n";
      }
      else if(tokens[0] == "T"){
        set_terminals(tokens);
        terminals_defined = true;
	//cerr << "Terminals defined\n";
      }
      else if(tokens[0] == "S"){
        set_start(tokens);
        start_found = true;
	//cerr << "Start defined\n";
      }
      else{
        useful_lines.push_back(current_line);
	//cerr << "Useful line found\n";
      }
    }
  }
  
  // check that terminals and nonterminals were defined
  if(!(nonterminals_defined && terminals_defined && start_found))
    throw ErrorExcept("Nonterminals, terminals and start expression must be defined");


  // set concatenation pointer for use in evaluating results
  if(element_map.find("Concat") != element_map.end()){
    concatOp = element_map["Concat"];
  }


  // construct production rules
  create_rules(useful_lines);
}

///
/// creates rules for production and places them in
/// map
/// @param useful_lines vector containing lines for producing rules
///
void Grammar::create_rules(vector<string> & useful_lines){
  vector<string> tokens;
  string rule_name;
  int rule_index = 0;
  vector<string>::iterator i = useful_lines.begin();

    do{
      tokens.clear();
      split_string(*i, " ", "\t", tokens);
      unsigned int j=0;
      for(j=0; j<tokens.size(); j++){
        if(tokens[j] == "::=" || tokens[j] == "|")
          break;
      } 
      if(tokens[j] == "::="){
        rule_index = 0;
        rule_name = tokens[1];
	vector <vector <Element*> > newmevve;
	rules[element_map[rule_name]] = newmevve;
        rules[element_map[rule_name]].push_back(vector<Element*>(0));
        j+=1;
      }
      else{
        rule_index++;
        rules[element_map[rule_name]].push_back(vector<Element*>(0)); 
        j+=1;
      }
      if(rule_name != "v"){
        // push back element pointers onto rule list
        for(;j<tokens.size(); j++){
          // ignore '<' and '>'
          if(tokens[j] == "<" || tokens[j] == ">")
            continue;
          // check that element exists
          map<string, Element* >::iterator k = element_map.find(tokens[j]);  

          // if k is missing check to see if this token is a number
          // if so, create a new number element for it and add to element map
          // otherwise throw exception
          if(k == element_map.end()){
            if(Stringmanip::isnumber(tokens[j]))
              element_map[tokens[j]] = terminal_machine.create_constant(tokens[j]);
            else
              throw ErrorExcept("ERROR: " + tokens[j] + " is not a valid element in this grammar");
          }
          rules[element_map[rule_name]][rule_index].push_back(element_map[tokens[j]]);
        }
      }
      else{ 
				// handle variable creation separately due to potentially very large number of variables
        // when find variable indicator create as many as indicated
        // need to split on position of hyphen
        int hyphenpos = tokens[j].find("-");
        // separate two numbers
        stringstream ss;
        string startstring = tokens[j].substr(1, hyphenpos-1);
        ss << startstring << " ";
        string endstring = tokens[j].substr(hyphenpos+1, tokens[j].length() - hyphenpos);
        ss << endstring;
        int startvar, endvar;
        ss >> startvar >> endvar;
        startvar -= 1;

        // now need to add each to the rule for 'v'
        for(int varnum=startvar; varnum < endvar; varnum++){
          stringstream varstream;
          varstream << varnum+1;
          string varstring = "V" + varstream.str();
          rules[element_map[rule_name]][rule_index].push_back(element_map[varstring]);
          rule_index++;
          if(varnum+1 < endvar)
            rules[element_map[rule_name]].push_back(vector<Element*>(0));
        }
      }
    }while(++i != useful_lines.end());  // go to next when    
}


///
/// Each nonterminal is created as an Element object
/// the pointers are stored in the map_names map
/// @param tokens vector contains symbols indentifying 
/// nonterminals in the grammar 
///
void Grammar::set_nonterminals(vector<string> & tokens){
  
  // '{' marks start of nonterminal list
  // '}' marks end of nonterminals
  unsigned int i;
  for(i=0; i<tokens.size(); i++){
    if(tokens[i] == "{"){
      i++;
      break;
    }
  }
  
  for(; i<tokens.size() && tokens[i] != "}"; i++){
    if(tokens[i] != ",")
      element_map[tokens[i]] = new Element(tokens[i], false);
  }
}

///
/// Sets start to be the nonterminal that begins each individual
/// @param tokens vector containing start symbol
///
void Grammar::set_start(vector<string> & tokens){
  unsigned int i;
  for(i=0; i<tokens.size(); i++){
      if(tokens[i] == "="){
        i++;
        break;
      }
  }
  start_codon = element_map[tokens[i]];
  //cerr << "Start codon set to " << start_codon << "\n";
}

///
/// Inserts spaces when needed around syntactical elements
/// this is needed for easy splitting of the grammar rules
/// @param line string to insert spaces into
///
void Grammar::insert_spaces(string & line){
  /*Make this a long so it works for large files... there should be a #define somewhere
    to make this be more automatic*/
  unsigned long int startsearch = 0, pos; 
  string searchstring = "()<>{}=,";
  // first search for ::= as it is a multiple character symbol
  pos = line.find("::=", startsearch);
  
  while(pos != string::npos){
    if((pos > 0) && line[pos-1] != ' ')
      line.insert(pos++, " ");
    if(pos+2 < line.length()  && line[pos+3] != ' '){
      line.insert(pos+3, " ");}
    startsearch = pos+3;
    pos =line.find("::=", startsearch);
    searchstring = "()<>{},";
  }

  // check for single character symbols and insert spaces around as needed  
  startsearch = pos = 0;

  pos = line.find_first_of(searchstring, startsearch);
  while(pos != string::npos){
    if((pos > 0) && line[pos-1] != ' ')
      line.insert(pos++, " ");
    if(pos < line.length() && line[pos+1] != ' ')
      line.insert(pos+1, " ");
    startsearch = pos+1;
    pos = line.find_first_of(searchstring, startsearch);        
  }
}

///
/// Retrieve start element
/// @return start element
///
Element * Grammar::get_start() const{
  return start_codon;
}

///
/// Retrieve left parentheses element
/// @return left parentheses element
///
Element * Grammar::get_left_paren(){
  return element_map["("];
}

///
/// Retrieve right parentheses element
/// @return right parentheses element
///
Element * Grammar::get_right_paren(){
  return element_map[")"];
}

///
/// Retrieve comma element
/// @return comma element
///
Element * Grammar::get_comma(){
  return element_map[","];
}

///
/// Retrieves the rules for a given element and codon value
/// @param codon_value selects the rules to return
/// @param el element that is being referenced
/// @return vector of Element pointers corresponding to the rules
/// selected
/// 
vector<Element *> Grammar::get_rules(int codon_value, Element * el){
  return rules[el][codon_value % rules[el].size()];
}


// split string into tokens as defined by the delimiter
int Grammar::split_string(const string& input, 
  string delim1, string delim2, vector<string>& results){
  int iPos = 0;
  int newPos = -1;
  string delimiter = delim1 + delim2;

  int sizeS2 = 1;
  int isize = input.size();

  vector<unsigned int> positions;
  newPos = input.find_first_of (delimiter, 0);

  if( newPos < 0 ) { return 0; }

  int numFound = 0;

  while( newPos >= iPos ){
    numFound++;
    positions.push_back(newPos);
    iPos = newPos;
    newPos = input.find_first_of (delimiter, iPos+sizeS2);
  }

  for(unsigned int i=0; i <= positions.size(); i++ ){
    string s;
    if( i == 0 ) {s = input.substr( i, positions[i] ); }

    /*So as to avoid negative array index:  causes crashes in some cases*/
    if (i > 0) {
      int offset = positions[i-1] + sizeS2;

      if( offset < isize )
	{
	  if( i == positions.size() )
	    {
	      s = input.substr(offset);
	    }
	  else if( i > 0 )
	    {
	      s = input.substr( positions[i-1] + sizeS2, 
				positions[i] - positions[i-1] - sizeS2 );
	    }
	}
    } /*End if i > 0*/

    if( s.size() > 0 )
    {
      results.push_back(s);
    }
  }
 
  return results.size();
}


///
/// each nonterminal is created as an Element object
/// the pointers are stored in the map_names map
/// @param tokens string tokens containing terminal elements
///
void Grammar::set_terminals(vector<string> & tokens){
  
  // '{' marks start of terminal list
  // '}' marks end of terminals
  unsigned int i;
  for(i=0; i<tokens.size(); i++){
    if(tokens[i] == "{"){
      i++;
      break;
    }
  }
  // alter to handle variables as V1-1000
  // instead of having to do each individually
  for(; i<tokens.size() && tokens[i] != "}"; i++){
    if(tokens[i] == ",")
      continue;
    if(tokens[i][0] == 'V' && tokens[i].find("-") != string::npos){
      // when find variable indicator create as many as indicated
      // need to split on position of hyphen
      int hyphenpos = tokens[i].find("-");
      // separate two numbers
      stringstream ss;
      string startstring = tokens[i].substr(1, hyphenpos-1);
      ss << startstring << " ";
      string endstring = tokens[i].substr(hyphenpos+1, tokens[i].length() - hyphenpos);
      ss << endstring;
      int startvar, endvar;
      ss >> startvar >> endvar;
      startvar -= 1;
      // need to create a token for each that is V#
      // so have V1, V2, etc.
      for(int varnum=startvar; varnum < endvar; varnum++){
        stringstream varstream;
        varstream << varnum+1;
        string varstring = "V" + varstream.str();
        element_map[varstring] = terminal_machine.create_terminal(varstring);
      }
    }
    else{
      element_map[tokens[i]]= terminal_machine.create_terminal(tokens[i]);
    }
  }

}

///
/// Determines which rules are recursive and which are not 
/// @param maxDepth 
///
void Grammar::calculate_rule_values(int maxDepth){
  Rule_status rStat;
  initTreeDepth = maxDepth;
  // initialize the map containing information on rules
  rStat.isrecursive=false;
  rStat.minSteps=0;
  map< Element*, vector<vector<Element*> > >::iterator rulesIter;
  
  // now establish whether every rule is recursive or not
  map<Element*, bool> searchElements;
  Element* currentElement;
  for(rulesIter=rules.begin(); rulesIter!=rules.end(); rulesIter++){
    rule_values[rulesIter->first] = vector<Rule_status>(rulesIter->second.size(), rStat);
    currentElement = rulesIter->first;
    

    for(unsigned int i=0; i<rules[currentElement].size(); i++){
      for(unsigned int j=0;j<rules[currentElement][i].size(); j++){
        if(rules[currentElement][i][j]->get_isterminal() == false){
          searchElements.clear();
          searchElements[currentElement] = true;
          if(check_recursion(rules[currentElement][i][j], searchElements)){
            rule_values[currentElement][i].isrecursive = true;
            recursiveNonTerms[rules[currentElement][i][j]] = true;
            // exit inner loop
            break;
          }
        }
      }
    }
  }
  
  // calculate the min depth for each rule in the grammar
  for(rulesIter=rules.begin(); rulesIter!=rules.end(); rulesIter++){
    currentElement = rulesIter->first;
    for(unsigned int i=0; i<rules[currentElement].size(); i++){
      int minDepth = 0;
      for(unsigned int j=0;j<rules[currentElement][i].size(); j++){
        int thisMinDepth = 0;
        if(rules[currentElement][i][j]->get_isterminal() == false){
          searchElements.clear();
          thisMinDepth = calc_element_min_depth(rules[currentElement][i][j], searchElements, 1);
        }
        else{
          thisMinDepth = 1;  // minDepth for a terminal is one
        }
        if(thisMinDepth > minDepth){
          minDepth = thisMinDepth;
        }
      }
      rule_values[currentElement][i].minSteps=minDepth; 
      
    }
  }

  // build rules map that contains a 2-D vector that lists
  // the valid rule indices for sensible initialization at any depth level
  for(rulesIter=rules.begin(); rulesIter!=rules.end(); rulesIter++){
    Element* currentElement = rulesIter->first;
	vector <vector <int> > new_vvi1;
	vector <vector <int> > new_vvi2;
	valid_rule_indices_all[currentElement] = new_vvi1;
	valid_rule_indices_recursive[currentElement] = new_vvi2;
    for(int currentDepth=0; currentDepth<=maxDepth; currentDepth++){
      valid_rule_indices_all[currentElement].push_back(vector <int> (0,0));
      valid_rule_indices_recursive[currentElement].push_back(vector <int> (0,0));
      for(unsigned int ruleIndex=0; ruleIndex<rules[currentElement].size(); ruleIndex++){
        if(rule_values[currentElement][ruleIndex].minSteps <= currentDepth){
          valid_rule_indices_all[currentElement][currentDepth].push_back(ruleIndex);
          if(rule_values[currentElement][ruleIndex].isrecursive == true){
            valid_rule_indices_recursive[currentElement][currentDepth].push_back(ruleIndex);
          }
        }
      }
    }
  }
}

///
/// returns number of valid recursive rules for particular depth
/// when conducting sensible initialization
/// @param ruleElement Element pointer 
/// @param depthLeft depth left in tree for rules
///
int Grammar::get_num_recursive_opts(Element* ruleElement, int depthLeft){
  return valid_rule_indices_recursive[ruleElement][depthLeft].size();
}

///
/// returns number of valid rules for particular depth
/// when conducting sensible initialization
/// @param ruleElement current Element
/// @param depthLeft depth left in tree for rules
///
int Grammar::get_num_opts(Element* ruleElement, int depthLeft){
  return valid_rule_indices_all[ruleElement][depthLeft].size();
}

///
/// returns actual rule number for random number
/// selected for recursive rules
/// @param ruleElement current Element
/// @param depthLeft depth left in tree for rules 
/// @param randIndex index selecting rule
///
int Grammar::get_real_recursive_index(Element* ruleElement, int depthLeft, int randIndex){
  return valid_rule_indices_recursive[ruleElement][depthLeft][randIndex];
}

///
/// returns number of valid rules for particular depth
/// when conducting sensible initialization
/// @param ruleElement current Element
/// @param depthLeft depth left in tree for rules
/// @param randIndex index selecting rule
///
int Grammar::get_real_index(Element* ruleElement, int depthLeft, int randIndex){
  return valid_rule_indices_all[ruleElement][depthLeft][randIndex];
}


///
/// recursive function to check whether an element is
/// recursive
/// @param startElement is element currently being checked to see if it
///       refers to the searchElement
/// @param searchElements are nonterminals that have already been used as searchElements
/// @return true if recursion found
///
bool Grammar::check_recursion(Element* startElement, map<Element *, bool> searchElements){
  
  // check to see if this element has been checked before
  // in this recursive search
  if(searchElements[startElement] == true){
    return true;
  }
  else{
    searchElements[startElement] = true;
  }
   
  for(unsigned int i=0; i<rules[startElement].size(); i++){
    for(unsigned int j=0; j<rules[startElement][i].size(); j++){
    // if the nonterminal matches one searched for or is stored as recursive then return true
      if((searchElements[rules[startElement][i][j]] == true) || 
        (recursiveNonTerms[rules[startElement][i][j]] == true)){  
        return true;
      }
      else if(rules[startElement][i][j]->get_isterminal() == false){
        // check any nonterminals 
        if(check_recursion(rules[startElement][i][j], searchElements)){
          return true;
        }
      }
    } 
  }
  
  //if function reaches this point searchElement has not been found
  return false;
}

///
/// Calculates and returns minimum tree depth for a rule to complete
/// @param currentElement
/// @param prevElements
/// @return minimum depth for the element 
///
int Grammar::calc_element_min_depth(Element* currentElement, map<Element*, bool> prevElements,
  int argDepth){
  int minDepth=10000000;
  // set this element to true in prevElements map 
  // to track which elements are recursive
  prevElements[currentElement] = true;
  
  // for each rule of this element check to see if the rule
  // has been established in the rule_values 
  for(unsigned int i=0; i<rules[currentElement].size(); i++){
    int thisDepth=0;
    if(rule_values[currentElement][i].minSteps){
      thisDepth = rule_values[currentElement][i].minSteps;
    }
    else{
      // when not established have to calculate the minDepth and 
      // track it if less than the current minimum
      bool ruleRecursive = false;
      for(unsigned int j=0; j<rules[currentElement][i].size(); j++){
        int depth = 0;
        if(rules[currentElement][i][j]->get_isterminal()){
          depth = 1;
        }
        else{
          // check to see if recursive
          if(prevElements[rules[currentElement][i][j]]==true){
            ruleRecursive=true;  
            break;
          }
          depth = calc_element_min_depth(rules[currentElement][i][j], prevElements, 1);
        }
        if(depth > thisDepth){
          thisDepth = depth;
        } 
      }
      if(!ruleRecursive){
        rule_values[currentElement][i].minSteps = thisDepth;
      }
      else{
        thisDepth = 10000000;
      }
    }

    if(thisDepth < minDepth){
      minDepth = thisDepth;
    }
  }
  return minDepth + argDepth;
}

///
/// Retrieves number of rules for any element
/// @param ruleElement Element for checking number of rules
/// @return number of rules for the element 
///
int Grammar::get_num_rules(Element* ruleElement){
  return rules[ruleElement].size();
}

///
/// Retrieves initial tree depth
/// @return tree depth
///
int Grammar::get_init_tree_depth() const{
  return initTreeDepth;
}

