// InputReader.cpp

#include "InputReader.h"
#include "mpi.h"
#include "unistd.h"

///
/// Reads input data file, records status of individuals in
/// a Tester, and records all variable values in Variable
/// @param inputfile Data file with variables and individual
/// status
/// @param testobj Tester for storing status and testing
/// neural networks that are created
///
void InputReader::read_input(string inputfile, Tester & testobj){
  
  cerr << "About to open input file\n";
  ifstream input(inputfile.c_str(), ios::in);
  if(!input.is_open()){
    throw ErrorExcept(inputfile + ":  cannot open!");
  }
  cerr << "Opened input file\n";

	if (memoryManager == mmdouble) {
		cerr << "Using MM mmdouble\n";
	}
	else if (memoryManager == mmgenotypeMiser) {
		cerr << "Using MM mmgenotypeMiser\n";
	}
	else {
		cerr << "Your MM a splode:  invalid MM\n";
	}
   
  char line[LINEMAX];
  int linecounter = 0;
  string current_line;
  Variable::clear_variable_list();
  testobj.clear_status();
  
  while(!input.eof()){
    linecounter++;
    input.getline(line, LINEMAX);
    
    string current_line(line);
    int linelength;
    if(!(linelength = current_line.length())) {
      continue; // blank line so go to next one in file    
		}
    vector<string> tokens;
    //split the string on either space or tab
    // last token will be the status of the individual (either 0 or 1)
    if(Stringmanip::split_string(current_line, " ", "\t", tokens) > 0){
      if(!(Stringmanip::isnumber(tokens.back()))) {
        cerr << "On line " << linecounter << ":\n";
        throw ErrorExcept("ERROR: " + tokens.back() + " is not a number in " + inputfile);
      }
	    
      //Slightly wasteful, but this memory is cheap...
      testobj.add_status_quan(Stringmanip::stodata(tokens.back()));
      testobj.add_status(Stringmanip::stoi(tokens.back()));  
    }
    
    vector<INPUT_TYPE> data; //old storage
		vector<BITTYPE> databits;  //new storage

		/*New magic - used for mmgenotypeMiser*/
		unsigned int num = 0;
		int shift;
		int varspernum = 2 * sizeof( BITTYPE ); /*2 4-bit fields in a byte*/
		
    for(unsigned int i=0; i<tokens.size()-1; i++){
      
			/*The old way of doing things*/
			if (memoryManager == mmdouble) {
				data.push_back(Stringmanip::stoinput(tokens[i]));
			}
			else if (memoryManager == mmgenotypeMiser) {

				//cerr << "InputReader::mmgenotypeMiser\n";

				/*The new way of doing things*/
				INPUT_TYPE tmp = Stringmanip::stoinput(tokens[i]);
				
				if ( (-1 != tmp) && (0 != tmp) && (1 != tmp) && (2 != tmp) &&
						 (3 != tmp) ) {
					cerr << "ERROR:  unexpected data value of " << tmp << " (" <<
						tokens[i] << ") encountered!\n";
					exit(1);
				}
				else {
					
					tmp++; /*ott-encoding starts from -1, our field starts from 0*/
					
					num = i / varspernum;
					shift = i % varspernum;
					
					//cerr << "Variable " << i << " has num= " << num << " and shift=" <<
					//	shift << "\n";
					
					if (0 == shift) {
						databits.push_back(tmp);
					}
					else {
						tmp = (tmp << (4 * shift));
						databits[num] = databits[num] | tmp;
					}
					
				} /*End bit-packing*/
			} /*End if mmGenotypeMiser*/
			else {
				cerr << "Attempting to temporarily store data in unspecified manner, exiting\n";
				//exit(1);
			}

    } /*End foreach genotype*/
		
		/*At this point, data has the raw data, and databits has bit-packed version*/

		/*Check to make sure data and databits have the same stuff*/
		/*
		BITTYPE tempvar;
		INPUT_TYPE allele;
		BITTYPE mask;
		for (unsigned int i = 0; i < data.size(); i++) {
			int index = i / varspernum;
			int shift = 4 * (i % varspernum);
			mask = 7;

			tempvar = databits[index];
			mask = (mask << shift);
			allele = tempvar & mask;
			allele = (allele >> shift);
			allele--;
			
			cerr << "Bit-packed allele: " << allele << " Vanilla variable: " <<
				data[i] << "\n";

				} */
		//exit(1);
		
		if (memoryManager == mmdouble) {
			Variable::fill_variable_list(data);
		}
		else if (memoryManager == mmgenotypeMiser) {
			//cerr << "InputReader::fill_variable_list_bits\n";
			Variable::fill_variable_list_bits(databits);
		}
		else {
			cerr << "Attempting to store data in unspecified manner, exiting\n";
			exit(1);
		}
  }
  input.close();
 
  cerr << "Finished reading input\n";

}

/*Sets which memory manager gets used, so we know how to store 
input data.*/
void InputReader::set_memory_manager(MemoryManagers memManager) {
	memoryManager = memManager;
}
