/* ----------------------------------------------------------------------------
  gennMPI.cpp

  Nicholas Hardison
  11-05-07

  SPMD MPI wrapper for gennMaster and gennSlave

  If rank is 0, call gennMaster.  Else, call gennSlave
	-----------------------------------------------------------------------------*/

#include "mpi.h"

#include "mpi.h"
#include <ga/GASimpleGA.h>
#include "GEBinStrBlockCrossGenome.h"
#include "GARandomSearchGA.h"
#include "genome.h"
#include "Grammar.h"
#include "Translator.h"
#include "Tester.h"
#include "InputReader.h"
#include "Config.h"
#include <iostream>
#include <string>
#include <limits.h>

#include <stdio.h>
#include "PVMDemeGA.h"
#include "genome.h"
#include "Config.h"
#include "gennMPI.h"

Translator trans;
Tester genn_test;
Grammar gram;

using namespace std;

//void output_model(GAGeneticAlgorithm& genome);
void output_population(const GAPopulation & pop, ostream & fileos);

double testing_set(string testfile, GEBinStrBlockCrossGenome & genome, Config config);
double calculate_error(GEBinStrBlockCrossGenome & genome);

/*NH:  this is experimental*/
/* namespace genn {
void output_model(GEBinStrBlockCrossGenome& bestgenome, ostream & output_file,
									bool output_notfinal = false); 
									} *//*End namespace genn*/

void output_errors(double train_error, double test_error, ostream & output_file);
void header(string progname, string versiondate);

int mainSlave(int argc, char ** argv);
int mainMaster(int argc, char ** argv);

const string program_name = "gennMPI";
const string version_date = "11/05/07";

int main(int argc, char ** argv) {

	int masterid = 0;
	int myrank = -1;

	MPI_Init(&argc,&argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &myrank);

	if (myrank == masterid) {
		mainMaster(argc, argv);
	}
	else {
		mainSlave(argc, argv);
	}

} /*End main*/


int mainSlave(int argc, char ** argv) {


  unsigned int randSeed;
	int masterid = 0;
	int mytid;
	MPI_Comm_rank(MPI_COMM_WORLD, &mytid);
	MPI_Status Stat;

	int status = MPI_Recv(&randSeed, 1, MPI_UNSIGNED, 0, MSG_SET_RANDOM_SEED,
					 MPI_COMM_WORLD, &Stat);

  GARandomSeed(randSeed);
  Config config;
  
  try{
    config.read_config(argv[1]);
  }
  catch(ErrorExcept & ee){
    cerr << endl << ee << endl << endl;
    exit(1);
  }
  InputReader iReader;
	iReader.set_memory_manager(config.get_memory_manager());
  
  string gramfile, testfile, trainfile;
	cerr << "Starting input reading\n";
  // read input files for grammar and data
  try{
    gram.read_gram_file(config.get_gramfile());
    if(config.get_init_depth()){
      gram.calculate_rule_values(config.get_init_depth());
    }
    iReader.read_input(config.get_trainfile(), genn_test);
    // set missing data value for the set
    Variable::set_missing_value(config.get_missing_value());
		Variable::set_memory_manager(config.get_memory_manager());

  }
  catch(ErrorExcept & ee){
    cerr << endl << ee << endl << endl;
    exit(1);
  }
  
  // set fitness parameter in Tester
  genn_test.set_fitness_type(config.get_fitness_type());
  
  // set codon size
  GEBinStrBlockCrossGenome::CODONSIZE = config.get_codon_size();

  // set translator parameters
	trans.set_codon_size(config.get_codon_size());
	trans.set_grammar(&gram);
	trans.set_wrap_count(config.get_wrapper_count());

  
  GEBinStrBlockCrossGenome genome(1*config.get_codon_size(),GenomeEvaluator);
  
  if(config.get_crossover_type() != standardCross){
    if(config.get_crossover_type() == matchCross){
      genome.crossover(genome.MatchStartCrossover);
    }
    else{
      genome.crossover(genome.BlockPreserveCrossover);
    }
  }
  
  genome.resizeBehaviour(config.get_min_chrom()*config.get_codon_size(), 
    config.get_max_chrom()*config.get_codon_size());


  // Use sensible initialization when called for 
  // in the config file
  if(config.get_init_depth()){
    genome.initializer(SensibleInitializer);
  }

  GAGeneticAlgorithm * gaPtr;
  GARandomSearchGA * gaRandomPtr = NULL;
  if(config.get_algorithm_type() == simpleAlgorithm){
    gaPtr = new GASimpleGA(genome);
  }
  else{
    gaRandomPtr = new GARandomSearchGA(genome);;
    gaPtr = gaRandomPtr; 
  }
  GAGeneticAlgorithm & ga = *gaPtr;
  
  // determine selection type and set
  SelectionTypes sType = config.get_selection_type();
  if(sType == uniformSelection){
    GAUniformSelector selector;
    ga.selector(selector);    
  }
  else if(sType == tournamentSelection){
    GATournamentSelector selector;
    ga.selector(selector);
  }
  else if(sType == rouletteSelection){
    GARouletteWheelSelector selector;
    ga.selector(selector);
  }
  else if(sType == rankSelection){
    GARankSelector selector;
    ga.selector(selector);
  }
  else{
    cerr << endl << "Selection type not specified" << endl << endl;
    exit(1);
  }

  ga.maximize();                // we want to maximize the objective function

  ga.scaling(GANoScaling());                // set the scaling method to our sharing


  ga.populationSize(config.get_pop_size());        // how many individuals in the population
  ga.pMutation(config.get_mutate());                // likelihood of mutating new offspring
  ga.pCrossover(config.get_cross()); 

  ga.initialize();

	int currentTotalReceived=0;  

  // open file for debug logging of populations
  ofstream debugstream;
  if(config.get_debug()){
    string outfilename = config.get_base_output() + "." + Stringmanip::itos(randSeed) + ".debug";
    if(config.get_algorithm_type() == simpleAlgorithm){
      debugstream.open(outfilename.c_str(), ios::out);
    }
    else{
      gaRandomPtr->debugOn = true;
      gaRandomPtr->debugStreamSet(outfilename);
    }
  }

  int done = 0;
	cerr << "Slave beginning loop...\n";
  while(!done){

		int rc = MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &Stat);
    int ival;
    int randseed;
		int nothing = 0;  /*int sink*/
    if(MPI_SUCCESS == rc) {
      int msgtag, tid;
			msgtag = Stat.MPI_TAG;
			tid = Stat.MPI_SOURCE;
			//cerr << "DEBUG:  gennSlave:  I (" << mytid << ") has a message ("
			//		 << msgtag << ") from " << tid << "\n";

      switch(msgtag) {
        case MSG_DONE:
		        done = 1;
	        break;

        case MSG_SET_POPULATION_SIZE:
					cerr << "DEBUG:  This looked like dead code...\n";
					exit(2);
	        break;

        case MSG_INITIALIZE:  
	        break;

        case MSG_STEP:
	        ival = 0;
					status = MPI_Recv(&ival, 1, MPI_INT, tid, msgtag, MPI_COMM_WORLD, &Stat);
					currentTotalReceived = 0;
	        for(int i=0; i<ival; i++){
	          ga.step();
	          if(config.get_debug()){
	            if(config.get_algorithm_type() == simpleAlgorithm){
	              debugstream << "STEP" << endl;
	              output_population(ga.population(), debugstream);
	            }
	          }
	        }
	        ival = ga.generation();
					status = MPI_Send(&ival, 1, MPI_INT, masterid, MSG_STEP_COMPLETE, MPI_COMM_WORLD);
	        break;

        case MSG_INCOMING_MIGRATION:
					RecvMigration(ga, tid);
	        break;

        case MSG_SEND_BEST: // send best individual to the master process
					status = MPI_Recv(&nothing, 1, MPI_INT, tid, msgtag, MPI_COMM_WORLD, &Stat);
          SendMigration(&masterid, ga, 1, 1);
          break;

        case MSG_SEND_MIGRATION:
	        {
						cerr << "DEBUG:  WARNING:  MSG_SEND_MIGRATION in gennSlave being ignored\n";
						exit(2);
	        } 
	        break;

        case MSG_SEND_POPULATION:
	        SendPopulation(masterid, ga.population());
	        break;
        
					/*This looks like dead code*/
        case MSG_SEND_STATISTICS:
					cerr << "DEBUG: WARNING:  told to send statistics... ignoring\n";
					exit(2);
	        break;
	        
	      case MSG_SET_RANDOM_SEED:
					status = MPI_Recv(&randseed, 1, MPI_INT, tid, msgtag, MPI_COMM_WORLD, &Stat);

	        GARandomSeed(randseed);
	        
	        break;
        case MSG_REMOVE_LOG:
          break;         
        default:
	        cerr << argv[0] << ": unknown msgtag: " << msgtag << "\n";
					cerr << "DEBUG: task " << mytid << " asploding on msgtag " 
							 << msgtag << " (from " << tid << ")\n";
					exit(1);
	        break;
      } /*end switch(msgtag)*/
    } /*End if MPI_Recv == MPI_SUCCESS*/
    else {
			cerr << argv[0] << ": error from MPI_Probe: " << rc << "\n";
    }

  } /*end while not done*/
  
  if(config.get_debug() && config.get_algorithm_type() == simpleAlgorithm){
    debugstream.close();
  }

	MPI_Finalize();
  delete gaPtr;
  exit(0);
}

/*Is this even used?
  Apparently not.  Leaving commented out as dead, will remove later. */
/*void output_model(GAGeneticAlgorithm& ga){
  GEBinStrBlockCrossGenome& genome = (GEBinStrBlockCrossGenome&) ga.population().best(0);
  
  trans.convert_codons(genome);

  vector<Element *> term_list;
  vector<Element *> constant_list;
  trans.create_terminals(term_list, constant_list);
  
  cerr << "MODEL:" << endl;
  
    vector<string> new_list;
  // try to make it pretty by taking out all the Concatenation operators
  for(unsigned int i=0; i<term_list.size(); i++){
    if(term_list[i]->get_name() == "Concat"){
      // next should be '(', concatenate until get to ')'
      // and then go back and remove last character
      i++;
      string replacement;
      while(term_list[++i]->get_name() != ")")
        replacement += term_list[i]->get_name();
      new_list.push_back(replacement.substr(0, replacement.length()-1));
    }
    else{
      new_list.push_back(term_list[i]->get_name());
    }
  }
  
  // when you get a W go indent results
  // unindent when you reach the end
  // use stack to mark beginning and end of the indentation
  for(unsigned int i=0; i<new_list.size(); i++){
    cerr << new_list[i] << " ";
  }
  cerr << endl; 
  
  // delete any constants created earlier
  for(unsigned int conCounter=0; conCounter<constant_list.size(); conCounter++){
      delete constant_list[conCounter];
  } 
}
*/


// For debugging outputs all information for all members 
// of the population
// Arg:  population to output
void output_population(const GAPopulation & pop, ostream & fileos){
  
  // need to output codons, score, and variables present
  for(int currInd=0; currInd < pop.size(); currInd++){
    fileos << "Individual #" << currInd+1 << endl;
    GEBinStrBlockCrossGenome& genome = (GEBinStrBlockCrossGenome&)pop.individual(currInd);
		trans.convert_codons(genome);

    vector<unsigned int> codons = trans.get_codon_values();
    for(unsigned int i=0; i<codons.size(); i++)
      fileos << codons[i] << " ";
    fileos << endl;
    
    vector<Element *> term_list;
    vector<Element *> constant_list;
    int effectiveSize = trans.create_terminals(term_list, constant_list);
    
    for(unsigned int i=0; i<term_list.size(); i++){
      if((term_list[i]->get_name())[0] == 'V'){
        fileos << term_list[i]->get_name() << " ";
      }
    }

    fileos << endl;
    fileos << "genome Size=" << codons.size() << "\teffective Size=" << effectiveSize << endl;
    fileos << "fitness=" << genome.score() << endl;
    // delete any constants created earlier
    for(unsigned int conCounter=0; conCounter<constant_list.size(); conCounter++){
      delete constant_list[conCounter];
    }
  }
}

int mainMaster(int argc, char** argv) {
  
  if(argc < 4){
    header(program_name, version_date);
    exit(1);
  }
  Config config;
  // read input files for grammar and data
  try{  
    config.read_config(argv[1]);
  }
  catch(ErrorExcept & ee){
    cerr << endl << ee << endl << endl;
    exit(1);
  }


  InputReader iReader;
	iReader.set_memory_manager(config.get_memory_manager());

  // set codon size
  GEBinStrBlockCrossGenome::CODONSIZE = config.get_codon_size();

  GEBinStrBlockCrossGenome genome(1*config.get_codon_size(), GenomeEvaluator);
  genome.resizeBehaviour(config.get_min_chrom()*config.get_codon_size(), 
    config.get_max_chrom()*config.get_codon_size());

  if(config.get_crossover_type() != standardCross){
    if(config.get_crossover_type() == matchCross)
      genome.crossover(genome.MatchStartCrossover);
    else
      genome.crossover(genome.BlockPreserveCrossover);
  }

  PVMDemeGA ga(genome);
	ga.set_trans(trans);

  // need to set number of populations
  // each population is a separate process
	int npop = Stringmanip::stoi(argv[2]);
  ga.nPopulations(npop);

	//spawn used to do this
	ga.setPopSize(npop);


	int nothing = 0;
	int flag = npop;

	//should merge this back into another spawn
	flag = npop;
	int seed = config.get_seed();
	for (unsigned int currProc=0; currProc < npop; currProc++) {
		int eachseed;
		if(((unsigned int)(INT_MAX - seed)) >= (unsigned int)currProc){
      eachseed = currProc + seed;
    }
    else{
      eachseed = currProc - (INT_MAX - seed);
    }

		MPI_Send(&eachseed, 1, MPI_UNSIGNED, currProc + 1, 
						 MSG_SET_RANDOM_SEED,
						 MPI_COMM_WORLD);
		cerr << "Set seed of " << currProc + 1 << " to " << eachseed << "\n";

	} /*End setting seeds*/

  cout << "initializing..." << endl;
  ga.initialize(config.get_seed());

  ga.set_pvm_exchange_gen(config.get_pvm_update_interval());
  int num_generations = config.get_gen() /config.get_pvm_update_interval();
  ga.nGenerations(num_generations);
  cout << "evolving..." << endl;
	int stepCount = 0;
  while(!ga.done()){
    cerr << "Starting step " << stepCount << endl; 
    ga.step();
		//cout << "gennMaster finished step " << stepCount << endl;
			stepCount++;
  }
	//cerr << "DEBUG:  gennMaster done stepping\n";
  ga.flushScores();
  GEBinStrBlockCrossGenome bestgenome(1, GenomeEvaluator);
  ga.collectBest(&bestgenome);
  
  // ------------------- P R O C E S S   G A   B E S T ----------------------
  // Run on training and test sets to produce error rates
 
  // read input files for grammar and data
  try{
    gram.read_gram_file(config.get_gramfile());
    iReader.read_input(config.get_trainfile(), genn_test);
    // set missing data value for the set
    Variable::set_missing_value(config.get_missing_value());
		Variable::set_memory_manager(config.get_memory_manager());
  }
  catch(ErrorExcept & ee){
    cerr << endl << ee << endl << endl;
    exit(1);
  }
  
  // set fitness parameter in Tester
  genn_test.set_fitness_type(config.get_fitness_type());
  
  // set translator parameters
	trans.set_codon_size(config.get_codon_size());
	trans.set_grammar(&gram);
	trans.set_wrap_count(config.get_wrapper_count());
 
  string outfilename = config.get_base_output() + ".out";
  ofstream outfile(outfilename.c_str(), ios::out);
  
  double train_error = calculate_error(bestgenome);
  
  // now run test using test set
  double test_error = testing_set(config.get_testfile(), bestgenome, config);
  
  output_errors(train_error, test_error, outfile);
 
	genn::output_model(bestgenome, outfile, trans);
  
  cout << "DONE " << endl;
	cerr << "DONE\n";

	/*Tell slaves to stop*/
	int status = 0;
	for (int i = 1; i <= npop; i++) {
		status = MPI_Send(&nothing, 1, MPI_INT, i, MSG_DONE,
											MPI_COMM_WORLD);
	}

  return 0;
}

double testing_set(string testfile, GEBinStrBlockCrossGenome & genome, Config config){
  InputReader iReader;
	iReader.set_memory_manager(config.get_memory_manager());
  
  try{
    iReader.read_input(testfile, genn_test);
  }
  catch(ErrorExcept & ee){
    cerr << endl << ee << endl << endl;
    exit(1);
  }
  return calculate_error(genome);
}

double calculate_error(GEBinStrBlockCrossGenome & genome){
  double total = 0.0;
	trans.convert_codons(genome);
  vector<Element *> term_list;
  vector<Element *> postfix_stack;
  vector<Element *> constant_list;
  if(trans.create_terminals(term_list, constant_list)){
    // convert to postfix stack
    postfix_stack=trans.create_postfix_stack(term_list);
    // need to test the genome
    total = genn_test.determine_error_rate(postfix_stack);
  }
  
  return total;
}

void output_errors(double train_error, double test_error, ostream & output_file){
  cout << "training error=" << train_error << endl;
  cout << "testing error=" << test_error << endl;
  output_file << "training error=" << train_error << endl;
  output_file << "testing error=" << test_error << endl;
}

/*Output the model formatted as a neural network.  If output_notfinal is
	false (the default), it is output both to cout and output_file.  If
	it is true, it gets printed to output_file (hopefully cerr) with a tag
	of "INTERMEDIATE_NETWORK:" instead of "MODEL:"
*/
void genn::output_model(GEBinStrBlockCrossGenome& genome, ostream & output_file, 
												Translator trans, bool output_notfinal){
  cerr << "foo1\n";
	trans.convert_codons(genome);

  cerr << "foo2\n";

  vector<Element *> term_list;
  vector<Element *> constant_list;
	trans.create_terminals(term_list, constant_list);

	cerr << "foo3\n";
  
	if (output_notfinal) {
		output_file << "INTERMEDIATE_NETWORK:" << endl;
	}
	else {
		cout << "MODEL:" << endl;
		output_file << "MODEL:" << endl;
	}
  
	cerr << "foo4\n";
	vector<string> new_list;
  // try to make it pretty by taking out all the Concatenation operators
  for(unsigned int i=0; i<term_list.size(); i++){
    if(term_list[i]->get_name() == "Concat"){
      // next should be '(', concatenate until get to ')'
      // and then go back and remove last character
      i++;
      string replacement;
      while(term_list[++i]->get_name() != ")")
        replacement += term_list[i]->get_name();
      new_list.push_back(replacement.substr(0, replacement.length()-1));
    }
    else{
      new_list.push_back(term_list[i]->get_name());
    }
  }
  cerr << "foo5\n";
  // when you get a W go indent results
  // unindent when you reach the end
  // use stack to mark beginning and end of the indentation
  for(unsigned int i=0; i<new_list.size(); i++){
		if (output_notfinal) {
			output_file << new_list[i] << " ";
		}
		else {
			cout << new_list[i] << " ";
			output_file << new_list[i] << " ";
		}
  }
	if (output_notfinal) {
		output_file << endl;
	}
	else {
		cout << endl;
		output_file << endl; 
	}
	
	cerr << "fooNaN\n";

}



// Outputs information on program to user
void header(string progname, string versiondate){
        cout << endl;
        cout << "---------------------------------------------------" << endl;
        cout << "       " << "Name:    " << progname <<  endl;
        cout << "       " << "Date:    " << versiondate <<  endl;
        cout << "       " << "Usage:   " << progname << " <config file> <number of processes> <slave exe>"<< endl << endl;
        cout << "       " << "Example: " << progname << " modelconfig.genn 10 /home/dudek/lab/genn/exe/gennSlave" << endl;
        cout << "---------------------------------------------------" << endl << endl;
}

