/* ----------------------------------------------------------------------------
  gennMaster.C
  
  Scott Dudek
  9-24-04
  
  adapted from example in galib245 examples for PVM parallelization

  mbwall 5dec95
  Copyright (c) 1995-1996 Massachusetts Institute of Technology

 DESCRIPTION:
   Example program to illustrate use of GAlib with PVM.  This example uses a
master-slave configuration to parallelize the genetic algorithm.  In this case,
the master controls the evolution by specifying migrations between populations
on various processes.
---------------------------------------------------------------------------- */
#include <string>
#include <stdio.h>
#include <iostream.h>
//#include <ga/GEBinStrGenome.h>
#include "PVMDemeGA.h"
#include "genome.h"
#include "Grammar.h"
#include "Translator.h"
#include "GEBinStrBlockCrossGenome.h"
#include "Tester.h"
#include "InputReader.h"
#include "Config.h"


using namespace std;

double testing_set(string testfile, GEBinStrBlockCrossGenome & genome);
double calculate_error(GEBinStrBlockCrossGenome & genome);
void output_model(GEBinStrBlockCrossGenome& bestgenome, ostream & output_file);
void output_errors(double train_error, double test_error, ostream & output_file);
void header(string progname, string versiondate);

const string program_name = "gennMaster";
const string version_date = "9/7/06";

Translator trans;
Tester genn_test;
Grammar gram;

int mainMaster(int argc, char** argv) {
  
  if(argc < 4){
    header(program_name, version_date);
    exit(1);
  }
  Config config;
  // read input files for grammar and data
  try{  
    config.read_config(argv[1]);
  }
  catch(ErrorExcept & ee){
    cerr << endl << ee << endl << endl;
    exit(1);
  }

	int masterid = 0;
	MPI_Init(&argc,&argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &masterid);

	cerr << "DEBUG:  masterid is " <<  masterid << "\n";

  InputReader iReader;

  // set codon size
  GEBinStrBlockCrossGenome::CODONSIZE = config.get_codon_size();

  GEBinStrBlockCrossGenome genome(1*config.get_codon_size(), GenomeEvaluator);
  genome.resizeBehaviour(config.get_min_chrom()*config.get_codon_size(), 
    config.get_max_chrom()*config.get_codon_size());

  if(config.get_crossover_type() != standardCross){
    if(config.get_crossover_type() == matchCross)
      genome.crossover(genome.MatchStartCrossover);
    else
      genome.crossover(genome.BlockPreserveCrossover);
  }

  PVMDemeGA ga(genome);

  // need to set number of populations
  // each population is a separate process
	int npop = Stringmanip::stoi(argv[2]);
  ga.nPopulations(npop);

	/*
  if(ga.spawn(argv[3], argv[1], config.get_seed()) < 0) exit(1);
  */

	//spawn used to do this
	ga.setPopSize(npop);


	//spawn used to do this too
	int flag = npop;
	int nothing = 0;
	MPI_Status Stat;
	while (flag > 0) {
		MPI_Recv(&nothing, 1, MPI_INT, flag, MSG_READY,
						 MPI_COMM_WORLD, &Stat);
		cerr << "DEBUG:  gennMaster got MSG_READY from " << flag << "\n";
		flag--;
	}

  cout << "initializing..." << endl;
  ga.initialize(config.get_seed());

  ga.set_pvm_exchange_gen(config.get_pvm_update_interval());
  int num_generations = config.get_gen() /config.get_pvm_update_interval();
  ga.nGenerations(num_generations);
  cout << "evolving..." << endl;
	int stepCount = 0;
  while(!ga.done()){
    cout << "gennMaster starting step " << stepCount << endl; 
    ga.step();
		cout << "gennMaster finished step " << stepCount << endl;
			stepCount++;
  }
	cerr << "DEBUG:  gennMaster done stepping\n";
  ga.flushScores();
  GEBinStrBlockCrossGenome bestgenome(1, GenomeEvaluator);
  ga.collectBest(&bestgenome);
  
  // ------------------- P R O C E S S   G A   B E S T ----------------------
  // Run on training and test sets to produce error rates
 
  // read input files for grammar and data
  try{
    gram.read_gram_file(config.get_gramfile());
    iReader.read_input(config.get_trainfile(), genn_test);
    // set missing data value for the set
    Variable::set_missing_value(config.get_missing_value());
  }
  catch(ErrorExcept & ee){
    cerr << endl << ee << endl << endl;
    exit(1);
  }
  
  // set fitness parameter in Tester
  genn_test.set_fitness_type(config.get_fitness_type());
  
  // set translator parameters
  trans.set_codon_size(config.get_codon_size());
  trans.set_grammar(&gram);
  trans.set_wrap_count(config.get_wrapper_count());
 
  string outfilename = config.get_base_output() + ".out";
  ofstream outfile(outfilename.c_str(), ios::out);
  
  double train_error = calculate_error(bestgenome);
  
  // now run test using test set
  double test_error = testing_set(config.get_testfile(), bestgenome);
  
  output_errors(train_error, test_error, outfile);
  
  output_model(bestgenome, outfile);
  
  cout << "DONE " << endl;

	/*Tell slaves to stop*/
	int status = 0;
	for (int i = 1; i <= npop; i++) {
		status = MPI_Send(&nothing, 1, MPI_INT, i, MSG_DONE,
											MPI_COMM_WORLD);
	}

  return 0;
}

double testing_set(string testfile, GEBinStrBlockCrossGenome & genome){
  InputReader iReader;
  
  try{
    iReader.read_input(testfile, genn_test);
  }
  catch(ErrorExcept & ee){
    cerr << endl << ee << endl << endl;
    exit(1);
  }
  return calculate_error(genome);
}

double calculate_error(GEBinStrBlockCrossGenome & genome){
  double total = 0.0;
  trans.convert_codons(genome);
  vector<Element *> term_list;
  vector<Element *> postfix_stack;
  vector<Element *> constant_list;
  if(trans.create_terminals(term_list, constant_list)){
    // convert to postfix stack
    postfix_stack=trans.create_postfix_stack(term_list);
    // need to test the genome
    total = genn_test.determine_error_rate(postfix_stack);
  }
  
  return total;
}

void output_errors(double train_error, double test_error, ostream & output_file){
  cout << "training error=" << train_error << endl;
  cout << "testing error=" << test_error << endl;
  output_file << "training error=" << train_error << endl;
  output_file << "testing error=" << test_error << endl;
}

void output_model(GEBinStrBlockCrossGenome& genome, ostream & output_file){
  
  trans.convert_codons(genome);

  vector<Element *> term_list;
  vector<Element *> constant_list;
  trans.create_terminals(term_list, constant_list);
  
  cout << "MODEL:" << endl;
  output_file << "MODEL:" << endl;
  
    vector<string> new_list;
  // try to make it pretty by taking out all the Concatenation operators
  for(unsigned int i=0; i<term_list.size(); i++){
    if(term_list[i]->get_name() == "Concat"){
      // next should be '(', concatenate until get to ')'
      // and then go back and remove last character
      i++;
      string replacement;
      while(term_list[++i]->get_name() != ")")
        replacement += term_list[i]->get_name();
      new_list.push_back(replacement.substr(0, replacement.length()-1));
    }
    else{
      new_list.push_back(term_list[i]->get_name());
    }
  }
  
  // when you get a W go indent results
  // unindent when you reach the end
  // use stack to mark beginning and end of the indentation
  for(unsigned int i=0; i<new_list.size(); i++){
    cout << new_list[i] << " ";
    output_file << new_list[i] << " ";
  }
  cout << endl;
  output_file << endl; 
}



// Outputs information on program to user
void header(string progname, string versiondate){
        cout << endl;
        cout << "---------------------------------------------------" << endl;
        cout << "       " << "Name:    " << progname <<  endl;
        cout << "       " << "Date:    " << versiondate <<  endl;
        cout << "       " << "Usage:   " << progname << " <config file> <number of processes> <slave exe>"<< endl << endl;
        cout << "       " << "Example: " << progname << " modelconfig.genn 10 /home/dudek/lab/genn/exe/gennSlave" << endl;
        cout << "---------------------------------------------------" << endl << endl;
}
