/* ----------------------------------------------------------------------------
  gennSlave.cpp
  
  Scott Dudek
  9-24-02
  
  Adapted from example in galib245 for PVM parallelization of library

-----------------------------------------------------------------------------
  gennSlave.C
  mbwall 5dec95
  Copyright (c) 1995-1996 Massachusetts Institute of Technology

 DESCRIPTION:
   Each slave evolves its own population.  Between steps it checks to see if
any messages have appeared for it.  After each step it sends a message that 
says that it has completed a generation.  A slave can receive two kinds of 
messages, one to send individuals (and where to send them), and another to
receive them (and from where to receive them).  The master controls the
migrations by telling the slaves when and where to migrate individuals.  
   To control the actual migration algorithm, you can modify the section of
code that merges the received individuals into the population.  You can also
determine which individuals will be selected to migrate to another population.
---------------------------------------------------------------------------- */
//#include <pvm3.h>
#include "mpi.h"
#include <ga/GASimpleGA.h>
//#include <ga/GEBinStrGenome.h>
#include "GEBinStrBlockCrossGenome.h"
#include "GARandomSearchGA.h"
#include "genome.h"

#include "Grammar.h"
#include "Translator.h"
#include "Tester.h"
#include "InputReader.h"
#include "Config.h"
#include <iostream>
#include <string>
#include <limits.h>

Translator trans;
Tester genn_test;
Grammar gram;

using namespace std;

void output_model(GAGeneticAlgorithm& genome);
void output_population(const GAPopulation & pop, ostream & fileos);

int mainSlave() {


  unsigned int randSeed = Stringmanip::stouint(argv[2]);


  GARandomSeed(randSeed);
  Config config;
  
  try{
    config.read_config(argv[1]);
  }
  catch(ErrorExcept & ee){
    cerr << endl << ee << endl << endl;
    exit(1);
  }
  InputReader iReader;
  
  string gramfile, testfile, trainfile;
	cerr << "Starting input reading\n";
  // read input files for grammar and data
  try{
    gram.read_gram_file(config.get_gramfile());
    if(config.get_init_depth()){
      gram.calculate_rule_values(config.get_init_depth());
    }
    iReader.read_input(config.get_trainfile(), genn_test);
    // set missing data value for the set
    Variable::set_missing_value(config.get_missing_value());

  }
  catch(ErrorExcept & ee){
    cerr << endl << ee << endl << endl;
    exit(1);
  }
  
  // set fitness parameter in Tester
  genn_test.set_fitness_type(config.get_fitness_type());
  
  // set codon size
  GEBinStrBlockCrossGenome::CODONSIZE = config.get_codon_size();

  // set translator parameters
  trans.set_codon_size(config.get_codon_size());
  trans.set_grammar(&gram);
  trans.set_wrap_count(config.get_wrapper_count());

  
  GEBinStrBlockCrossGenome genome(1*config.get_codon_size(),GenomeEvaluator);
  
  if(config.get_crossover_type() != standardCross){
    if(config.get_crossover_type() == matchCross){
      genome.crossover(genome.MatchStartCrossover);
    }
    else{
      genome.crossover(genome.BlockPreserveCrossover);
    }
  }
  
  genome.resizeBehaviour(config.get_min_chrom()*config.get_codon_size(), 
    config.get_max_chrom()*config.get_codon_size());


  // Use sensible initialization when called for 
  // in the config file
  if(config.get_init_depth()){
    genome.initializer(SensibleInitializer);
  }

  GAGeneticAlgorithm * gaPtr;
  GARandomSearchGA * gaRandomPtr = NULL;
  if(config.get_algorithm_type() == simpleAlgorithm){
    gaPtr = new GASimpleGA(genome);
  }
  else{
    gaRandomPtr = new GARandomSearchGA(genome);;
    gaPtr = gaRandomPtr; 
  }
  GAGeneticAlgorithm & ga = *gaPtr;
  
  // determine selection type and set
  SelectionTypes sType = config.get_selection_type();
  if(sType == uniformSelection){
    GAUniformSelector selector;
    ga.selector(selector);    
  }
  else if(sType == tournamentSelection){
    GATournamentSelector selector;
    ga.selector(selector);
  }
  else if(sType == rouletteSelection){
    GARouletteWheelSelector selector;
    ga.selector(selector);
  }
  else if(sType == rankSelection){
    GARankSelector selector;
    ga.selector(selector);
  }
  else{
    cerr << endl << "Selection type not specified" << endl << endl;
    exit(1);
  }

  ga.maximize();                // we want to maximize the objective function

  ga.scaling(GANoScaling());                // set the scaling method to our sharing


  ga.populationSize(config.get_pop_size());        // how many individuals in the population
  ga.pMutation(config.get_mutate());                // likelihood of mutating new offspring
  ga.pCrossover(config.get_cross()); 

  ga.initialize();

	/*
  status = pvm_initsend(PvmDataDefault);
  status = pvm_send(masterid, MSG_READY);
	*/

	int currentTotalReceived=0;  

  // open file for debug logging of populations
  ofstream debugstream;
  if(config.get_debug()){
    string outfilename = config.get_base_output() + "." + Stringmanip::itos(randSeed) + ".debug";
    if(config.get_algorithm_type() == simpleAlgorithm){
      debugstream.open(outfilename.c_str(), ios::out);
    }
    else{
      gaRandomPtr->debugOn = true;
      gaRandomPtr->debugStreamSet(outfilename);
    }
  }

	int nothing = 0;
	status = MPI_Send(&nothing, 1, MPI_INT, masterid, MSG_READY,
										MPI_COMM_WORLD);

  int done = 0;
  while(!done){

    //int bufid = pvm_recv(-1, -1);
		MPI_Status Stat;
		//int flag = -1;

		int rc = MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &Stat);
    int ival;
    int randseed;
		int nothing = 0;  /*int sink*/
    if(MPI_SUCCESS == rc) {
      int msgtag, tid;
      //status = pvm_bufinfo(bufid, &bytes, &msgtag, &tid);
			msgtag = Stat.MPI_TAG;
			tid = Stat.MPI_SOURCE;
			//cerr << "DEBUG:  gennSlave:  I (" << mytid << ") has a message ("
			//		 << msgtag << ") from " << tid << "\n";

      switch(msgtag) {
        case MSG_DONE:
		        done = 1;
	        break;

        case MSG_SET_POPULATION_SIZE:
					cerr << "DEBUG:  This looked like dead code...\n";
					exit(2);
	        /*ival = gaDefPopSize;
	        status = pvm_upkint(&ival, 1, 1);
	        ga.populationSize(ival);*/
	        break;

        case MSG_INITIALIZE:  
	        break;

        case MSG_STEP:
	        ival = 0;
	        //status = pvm_upkint(&ival, 1, 1);
					status = MPI_Recv(&ival, 1, MPI_INT, tid, msgtag, MPI_COMM_WORLD, &Stat);
					//cerr << "DEBUG:  Slave " << mytid << " ack step " <<
					//	ival << " times\n";
					currentTotalReceived = 0;
	        for(int i=0; i<ival; i++){
	          ga.step();
	          if(config.get_debug()){
	            if(config.get_algorithm_type() == simpleAlgorithm){
	              debugstream << "STEP" << endl;
	              output_population(ga.population(), debugstream);
	            }
	          }
	        }
	        ival = ga.generation();
					/*
	        status = pvm_initsend(PvmDataDefault);
	        status = pvm_pkint(&ival, 1, 1);
	        status = pvm_send(masterid, MSG_STEP_COMPLETE);
					*/
					status = MPI_Send(&ival, 1, MPI_INT, masterid, MSG_STEP_COMPLETE, MPI_COMM_WORLD);
					//cerr <<"DEBUG:  Slave " << mytid << " sent " << masterid << 
					//	"msg_step_complete (" << ival << "), rc=" << status << "\n";
	        break;

        case MSG_INCOMING_MIGRATION:
					RecvMigration(ga, tid);
         
        
	        // altered to send ready after processing migrants
					/* removed nihard */
					/*
	        status = pvm_initsend(PvmDataDefault);
          status = pvm_send(masterid, MSG_READY);
					*/
	        
	        break;

        case MSG_SEND_BEST: // send best individual to the master process
					status = MPI_Recv(&nothing, 1, MPI_INT, tid, msgtag, MPI_COMM_WORLD, &Stat);
          SendMigration(&masterid, ga, 1, 1);
					//cerr << "DEBUG:  slave " << mytid << " sent best to " << 
					//	masterid << "\n";
          break;

        case MSG_SEND_MIGRATION:
	        {
						cerr << "DEBUG:  WARNING:  MSG_SEND_MIGRATION in gennSlave being ignored\n";
						exit(2);
						/*
            int numMigrateID, count;
            status = pvm_upkint(&numMigrateID, 1, 1);
            int * demeID = new int[numMigrateID];
            status = pvm_upkint(demeID, numMigrateID, 1);
            status = pvm_upkint(&count, 1, 1);
            SendMigration(demeID, ga, count, numMigrateID);
	          delete [] demeID;  
						*/
	        } 
	        break;

        case MSG_SEND_POPULATION:
	        SendPopulation(masterid, ga.population());
	        break;
        
					/*This looks like dead code*/
        case MSG_SEND_STATISTICS:
	        /*SendStatistics(masterid, ga.statistics());*/
					cerr << "DEBUG: WARNING:  told to send statistics... ignoring\n";
					exit(2);
	        break;
	        
	      case MSG_SET_RANDOM_SEED:
	        //status = pvm_upkint(&randseed, 1, 1);
					status = MPI_Recv(&randseed, 1, MPI_INT, tid, msgtag, MPI_COMM_WORLD, &Stat);

	        GARandomSeed(randseed);
	        
					/*
					status = pvm_initsend(PvmDataDefault);
          status = pvm_send(masterid, MSG_READY);
					*/
	        break;
        case MSG_REMOVE_LOG:
          break;         
        default:
	        cerr << argv[0] << ": unknown msgtag: " << msgtag << "\n";
					cerr << "DEBUG: task " << mytid << " asploding on msgtag " 
							 << msgtag << " (from " << tid << ")\n";
					exit(1);
	        break;
      } /*end switch(msgtag)*/
    } /*End if MPI_Recv == MPI_SUCCESS*/
    else {
			cerr << argv[0] << ": error from MPI_Probe: " << rc << "\n";
    }

  } /*end while not done*/
  
  if(config.get_debug() && config.get_algorithm_type() == simpleAlgorithm){
    debugstream.close();
  }
  //pvm_exit();

	cerr << "gennSlave done:  exiting.\n";
	MPI_Finalize();
  delete gaPtr;
  exit(0);
}


void output_model(GAGeneticAlgorithm& ga){
  GEBinStrBlockCrossGenome& genome = (GEBinStrBlockCrossGenome&) ga.population().best(0);
  
  trans.convert_codons(genome);

  vector<Element *> term_list;
  vector<Element *> constant_list;
  trans.create_terminals(term_list, constant_list);
  
  cerr << "MODEL:" << endl;
  
    vector<string> new_list;
  // try to make it pretty by taking out all the Concatenation operators
  for(unsigned int i=0; i<term_list.size(); i++){
    if(term_list[i]->get_name() == "Concat"){
      // next should be '(', concatenate until get to ')'
      // and then go back and remove last character
      i++;
      string replacement;
      while(term_list[++i]->get_name() != ")")
        replacement += term_list[i]->get_name();
      new_list.push_back(replacement.substr(0, replacement.length()-1));
    }
    else{
      new_list.push_back(term_list[i]->get_name());
    }
  }
  
  // when you get a W go indent results
  // unindent when you reach the end
  // use stack to mark beginning and end of the indentation
  for(unsigned int i=0; i<new_list.size(); i++){
    cerr << new_list[i] << " ";
  }
  cerr << endl; 
  
  // delete any constants created earlier
  for(unsigned int conCounter=0; conCounter<constant_list.size(); conCounter++){
      delete constant_list[conCounter];
  } 
}



// For debugging outputs all information for all members 
// of the population
// Arg:  population to output
void output_population(const GAPopulation & pop, ostream & fileos){
  
  // need to output codons, score, and variables present
  for(int currInd=0; currInd < pop.size(); currInd++){
    fileos << "Individual #" << currInd+1 << endl;
    GEBinStrBlockCrossGenome& genome = (GEBinStrBlockCrossGenome&)pop.individual(currInd);
    trans.convert_codons(genome);

    vector<unsigned int> codons = trans.get_codon_values();
    for(unsigned int i=0; i<codons.size(); i++)
      fileos << codons[i] << " ";
    fileos << endl;
    
    vector<Element *> term_list;
    vector<Element *> constant_list;
    int effectiveSize = trans.create_terminals(term_list, constant_list);
    
    for(unsigned int i=0; i<term_list.size(); i++){
      if((term_list[i]->get_name())[0] == 'V'){
        fileos << term_list[i]->get_name() << " ";
      }
    }

    fileos << endl;
    fileos << "genome Size=" << codons.size() << "\teffective Size=" << effectiveSize << endl;
    fileos << "fitness=" << genome.score() << endl;
    // delete any constants created earlier
    for(unsigned int conCounter=0; conCounter<constant_list.size(); conCounter++){
      delete constant_list[conCounter];
    }
  }
}
