#!/usr/bin/perl
#
# CrossValidation.pm - Bill White - 2/19/02
#
# This module defines the interface to a generic framework (API) for 
# performing 10-fold cross validation of pattern recognition methods. 
# Particular methods are defined in separate modules. This module also 
# defines several convenience subroutines for logging, formatting, etc.

use Time::localtime;

# bring in the method modules
# require "/home/bwhite/moore_lab_shares/GPNNCrossValidation.pm";
require "/home/nicholas/disttest/genndist/lib/GPCrossValidation.pm";
#require "/home/bwhite/moore_lab_shares/NICOCrossValidation.pm";
require "/home/nicholas/disttest/genndist/lib/GPNNCrossValidation.pm";
require "/home/nicholas/disttest/genndist/lib/GENNCrossValidation.pm";

my $numCases = 0;
my $numControls =0;
@trainCounts = ();
@testCounts = ();

##############################################################################
# Subroutine: ReportStdout - report a message to stdout
#
# Accepts: Message to report
# Returns: Nothing
##############################################################################
sub ReportStdout
{
	my $message = shift;
	my $tm = localtime;
	printf 
		"[%04d-%02d-%02d %02d:%02d:%02d] -> %s\n", 
		$tm->year+1900, 
		($tm->mon)+1, 
		$tm->mday,
		$tm->hour,
		$tm->min,
		$tm->sec,
		$message;
}

##############################################################################
# Subroutine: ReportFatalError - report a fatal error message to stdout 
#                                and exit
#
# Accepts: Message to report
# Returns: Nothing
#
# NOTE: as a side effect, this subroutine exits the running program!
##############################################################################
sub ReportFatalError
{
	my $message = shift;
	ReportStdout("FATAL ERROR: $message");
	exit;
}

##############################################################################
# Subroutine: ReadCVConfig - Read and parse configuration file into 
#                            configuration hash key->value pairs
#
# Accepts: Configuration filename
# Returns: Nothing
##############################################################################
sub ReadCVConfig
{
  my $configFile = shift;
	%newHash;
  my @configPair;
	
  open(CONFIG_FILE, "<$configFile") or 
		ReportFatalError("ReadCVConfig: Cannot open cross validation " . 
										 "config file: $configFile, $!");
  while(<CONFIG_FILE>) {
    chomp;

		# skip blank lines and comments
		if(($_ eq '') or (substr($_, 0, 1) eq '#')) {
			next;
		}

		# split the line on equals sign, trim spaces and store in global 
		# configuartion hash
    @configPair = split /=/;
		$configPair[0] =~ s/\s+//g;
		$configPair[1] =~ s/\s+//g;
		$newHash{$configPair[0]} = $configPair[1];
  }
  close(IFILE);

	# check for minimum configuration
	if(!defined($newHash{workingDirectory})) {
		ReportFatalError("ReadCVConfig: \"workingDirectory\" must be defined " . 
										 "in $configFile")
	}
	if(!defined($newHash{exeDirectory})) {
		ReportFatalError("ReadCVConfig: \"exeDirectory\" must be defined " . 
										 "in $configFile")
	}
	if(!defined($newHash{baseDatasetDirectory})) {
		ReportFatalError("ReadCVConfig: \"baseDatasetDirectory\" must be " . 
										 "defined in $configFile")
	}
	if(!defined($newHash{baseDatasetName})) {
		ReportFatalError("ReadCVConfig: \"baseDatasetName\" must be defined " . 
										 "in $configFile")
	}
	if(!defined($newHash{methodType})) {
		ReportFatalError("ReadCVConfig: \"methodType\" must be defined " . 
										 "in $configFile")
	}
	if(!defined($newHash{methodConfigTemplate})) {
		ReportFatalError("ReadCVConfig: \"methodConfigTemplate\" must be " . 
										 "defined in $configFile")
	}
	if(!defined($newHash{randomSeed})) {
		ReportFatalError("ReadCVConfig: \"randomSeed\" must be defined " . 
										 "in $configFile")
	}

	return %newHash;
}

##############################################################################
# Subroutine: DumpHash - Dump a hash to stdout
#
# Accepts: Hash, message
# Returns: Nothing
##############################################################################
sub DumpHash
{
	my $configHashRef = shift;
	my $message = shift;
	ReportStdout($message);
	my ($key, $value);
	while(($key, $value) = each(%$configHashRef)) {
		ReportStdout("$key => $value");
	}
}

##############################################################################
# Subroutine: WriteMethodConfigFiles - Write a set of config files for the
#                                      method being used for training and 
#                                      testing
# Accepts: Hash of configuration info
# Returns: Nothing
##############################################################################
sub WriteMethodConfigFiles
{
	my $configHashRef = shift;
	my $curCV;
	my $numCrossVals = $configHashRef->{crossValidation} || 10;
#	for($curCV=1; $curCV <= 10; $curCV++) {
	for($curCV=1; $curCV <= $numCrossVals; $curCV++) {
		if($configHashRef->{methodType} eq 'GPNN') {
			ReportStdout("Writing GPNN CV config file $curCV");
			WriteGPNNTrainConfigFile($curCV, $configHashRef);
			WriteGPNNTestConfigFile($curCV, $configHashRef);
		}
    elsif($configHashRef->{methodType} eq 'GP') {
      ReportStdout("Writing GP CV config file $curCV");
      WriteGPTrainConfigFile($curCV, $configHashRef);
      WriteGPTestConfigFile($curCV, $configHashRef);
    } 
    elsif($configHashRef->{methodType} eq 'GENN'){
      ReportStdout("Writing GENN CV config file $curCV");
      if($curCV==1){
        WriteGENNGrammarFile($configHashRef);
      }
      WriteGENNConfigFile($curCV, $configHashRef);
    }
		elsif($configHashRef->{methodType} eq 'NICO') {
			ReportStdout("Writing NICO CV config file $curCV");
			WriteNICOTrainConfigFile($curCV, $configHashRef);
			WriteNICOTestConfigFile($curCV, $configHashRef);
		}
		elsif($configHashRef->{methodType} eq 'ACO') {
			ReportFatalError("WriteMethodConfigFiles: ACO Not Implemented");
		}
		else {
			ReportFatalError("WriteMethodConfigFiles: " . 
											 $configHash{methodType} .
											 " Not Implemented");
		}
	}
}

##############################################################################
# Subroutine: GetDatasetInfo - Get characteristics of the dataset into hash
#
# Accepts: Hash of configuration info
# Returns: Nothing
##############################################################################
sub GetDatasetInfo
{
	my $configHashRef = shift;
	my $datasetDir = $configHashRef->{baseDatasetDirectory};
	my $dataset = $configHashRef->{baseDatasetName};
	my ($numRows, $numCols) = 
		GetDatasetCharacteristics("${datasetDir}/${dataset}");
	$configHashRef->{datasetRows} = $numRows;
	$configHashRef->{datasetCols} = $numCols;

	############################################################################
	# This is the section that Marylyn Ritchie modified.  Here we will
	# count the number of # cases and controls in the dataset so that we
	# can determine if we have balanced data or# not.  This will determine
	# how the splitting of the data takes place for cross # validation.  #
	############################################################################

	open(INFILE,"<${datasetDir}/${dataset}") or 
		die "Can't read from '${datasetDir}/${dataset}': $!";
 	while ($line = <INFILE>) {
		my @Genotypes = split/ /,$line;
		if ($Genotypes[$#numCols] == 1) {
			$numCases++;
		} else {
			$numControls++;
		}
	}	
	printf "The number of cases: $numCases. Number of controls: $numControls\n";
}



#############################################################################
# Subroutine SplitTileDataset - splits dataset into a number of percentile
# sets and the testing set is randomly selected from the individuals in 
# that percentile
# All status values need to be between 0 and 1.
# 
# Accepts: Hash of configuration info
# Returns: Nothing
#############################################################################
sub SplitTileDataset{

        my $configHashRef = shift;

        my $datasetDirectory = $configHashRef->{baseDatasetDirectory};
        my $datasetFilename  = $configHashRef->{baseDatasetName};
        my $originalDataset  = "${datasetDirectory}/${datasetFilename}";

        my $cwd = $configHashRef->{workingDirectory};

        my $numIndividuals = $configHashRef->{datasetRows};
        my $sortField = $configHashRef->{datasetCols};

	# number of individuals will equal numCases + numControls
	my $totalInds = $numCases + $numControls;
	
       # sort individuals into the proper bins
       my $numBins = $configHashRef->{numBins};
       # establish cutoff
       my $range = 1.0/$numBins;
      
       my @binRanges;
       my $endrange = 0;
       for(my $i=0; $i<$numBins-1; $i++){
         $endrange += $range; 
         push(@binRanges, $endrange);
       } 
       # add final end range
       push(@binRanges, 1.0);

       my @indbins;

     #  SystemCall("cp  ${cwd}/${datasetFilename}");
#	open(DATASET_FILE, "<${cwd}/${datasetFilename}") or
#                        ReportFatalError("Cannot open dataset file for reading: " .
#                                                                                         "${cwd}/${datasetFilename}: $!");

open(DATASET_FILE,  $originalDataset) or ReportFatalError("Cannot open dataset file: " .
$originalDataset);
                while (<DATASET_FILE>) {
                        chomp;
                        my @individualData = split" ";
                        for(my $i=0; $i<@binRanges; $i++){
		          if($individualData[$#individualData]<=$binRanges[$i]){
			    push(@{$indbins[$i]}, $_);
                            last;
			  }
                        } 
                }
                close(DATASET_FILE);

        # shuffle all the arrays
        for(my $i=0; $i<@indbins; $i++){
          FisherYatesShuffle($indbins[$i]); 
        }

      # distribute into the bins as evenly as possible
      @testbins;
      my $numCrossVals = $configHashRef->{crossValidation} || 10; 
      my $empty = 0;
      while(!$empty){
        $empty = 1;
        for(my $currcv = 0; $currcv<$numCrossVals; $currcv++){
          for(my $currbin=0; $currbin<@indbins; $currbin++){
            if(@{$indbins[$currbin]}>0){
              push(@{$testbins[$currcv][$currbin]}, shift(@{$indbins[$currbin]}));
              $empty=0;
            } 
          }
        } 
      }
      
      # with testing bins set output the training and testing files
      # for each cross-validation interval

      for(my $curCVFile=1; $curCVFile<=$numCrossVals; $curCVFile++){
        open(TRN_CV_FILE, ">${cwd}/${datasetFilename}.${curCVFile}.trn") or
            ReportFatalError("Cannot open dataset for writing: " .
                           "${cwd}/${datasetFilename}.${curCVFile}.trn: $!");
        open(TST_CV_FILE, ">${cwd}/${datasetFilename}.${curCVFile}.tst") or 
            ReportFatalError("Cannot open dataset for writing: " .
                             "${cwd}/${datasetFilename}.${curCVFile}.tst: $!");      
         
        for(my $currCV=1; $currCV<=$numCrossVals; $currCV++){
          for(my $currbin=1; $currbin<=@indbins; $currbin++){
            if($curCVFile != $currCV){
              foreach my $ind (@{$testbins[$currCV-1][$currbin-1]}){
                print TRN_CV_FILE "$ind\n";
              }  
            }
            else{
              foreach my $ind(@{$testbins[$currCV-1][$currbin-1]}){
                print TST_CV_FILE "$ind\n";
              }
            }
          }
        }

        close(TRN_CV_FILE);
        close(TST_CV_FILE);
      }
         
}


##############################################################################
# Subroutine: SplitDataset - Split dataset into cross validation sets
#
# Accepts: Hash of configuration info
# Returns: Nothing
##############################################################################
sub SplitDataset
{
		print "CrossValidation.pm SplitDataset called\n";
	my $configHashRef = shift;

	my $datasetDirectory = $configHashRef->{baseDatasetDirectory};
	my $datasetFilename  = $configHashRef->{baseDatasetName};
	my $originalDataset  = "${datasetDirectory}/${datasetFilename}";

	my $cwd = $configHashRef->{workingDirectory};

	my $numIndividuals = $configHashRef->{datasetRows};
	my $sortField = $configHashRef->{datasetCols};
	
	if($numCases != $numControls)  {
		printf "$originalDataset\n";
		my @params = ($originalDataset, $numCases, $numControls, $datasetFilename);
    print "UnbalancedDataSplit\n";
		UnbalancedDataSplit($configHashRef,@params);
	}
	else  {
    
    # default to 10 if no cross validation specified
    my $numCrossVals = $configHashRef->{crossValidation} || 10;

		my $numIndividualsDiv2 = int(($numIndividuals / 2));
#		my $testCVSize  = int($numIndividuals / 10);
    my $testCVSize = int($numIndividuals / $numCrossVals);
		my $trainCVSize = $numIndividuals - $testCVSize;
		
		ReportStdout("Train CV size: $trainCVSize");
		ReportStdout("Test CV size:  $testCVSize");

		# sort the file before reading it in
		ReportStdout("Dataset sort");
		SystemCall("sort -r -k $sortField $originalDataset > " . 
							 "${cwd}/${datasetFilename}.sorted");
	
		# read sorted dataset into array
		ReportStdout("Reading ${cwd}/${datasetFilename}.sorted into memory");
		my @individualData;
		open(DATASET_FILE, "<${cwd}/${datasetFilename}.sorted") or 
			ReportFatalError("Cannot open dataset file for reading: " . 
											 "${cwd}/${datasetFilename}.sorted: $!");
		while (<DATASET_FILE>) {
			chomp;
			push @individualData, $_;
		}
		close(DATASET_FILE);
		unlink "${cwd}/${datasetFilename}.sorted";

		# shuffle each group by shuffling array indices
		ReportStdout("Shuffling");
		my @indexArray1 = (0..($numIndividualsDiv2 - 1));
		my @indexArray2 = ($numIndividualsDiv2..($numIndividuals - 1));
		FisherYatesShuffle(\@indexArray1);
		FisherYatesShuffle(\@indexArray2);

		my $index = 0;
		my $curIndex;
		my @shuffledArray;
		foreach $curIndex (@indexArray1) {
			$shuffledArray[$index++] = $individualData[$curIndex];
		}
		foreach $curIndex (@indexArray2) {
			$shuffledArray[$index++] = $individualData[$curIndex];
		}

		# alternate affected/unaffected
		ReportStdout("Alternating");
		$index = 0;
		my @alternatedArray;
		for ($i=0; $i < $numIndividualsDiv2; $i++) {
			$alternatedArray[$index++] = $shuffledArray[$i];
			$alternatedArray[$index++] = $shuffledArray[$i + $numIndividualsDiv2];
		}

		# create CV training and testing file indices
		ReportStdout("Creating CV indices");
		my @dataFileIndices = (0..($numIndividuals - 1));
		my @trainingCVIndices;
		my @testingCVIndices;
		my ($beginLeaveOutIndex, $endLeaveOutIndex);
		my ($curTestingCount, $curTrainingCount);
		my ($curDataIndex, $curCV, $curInd);
#		for ($curCV=0; $curCV < 10; $curCV++) {
    for ($curCV=0; $curCV < $numCrossVals; $curCV++) {
			$trainCounts[$curCV+1] = $trainCVSize;
			$testCounts[$curCV+1] = $testCVSize;

			$beginLeaveOutIndex = $curCV * $testCVSize;
			$endLeaveOutIndex = $beginLeaveOutIndex + $testCVSize - 1;
			$curTestingCount = $curTrainingCount = 0;
			$curDataIndex = $curCV * $testCVSize;
			for ($curInd=0; $curInd < $numIndividuals; $curInd++) {
				if (($curInd >= $beginLeaveOutIndex) && 
						($curInd <= $endLeaveOutIndex)) {
					$testingCVIndices[$curCV][$curTestingCount++] = 
						$dataFileIndices[$curInd];
				} else {
					$trainingCVIndices[$curCV][$curTrainingCount++] = 
						$dataFileIndices[$curInd];
				}
				$curDataIndex++;
				if ($curDataIndex == $numIndividuals) {
					$curDataIndex = 0;
				}
			}

			# make sure the data loading counts match what we calculated
			if (($curTrainingCount != $trainCVSize) || 
					($curTestingCount != $testCVSize)) {
				ReportFatalError( "Calculated CV counts don't match loaded counts. " .
												 "Training: [$curTrainingCount] vs. [$trainCVSize], " .
												 "Testing: [$curTestingCount] vs. [$testCVSize]");
			}
		}

		# write all CV datasets
		ReportStdout("Writing CV files");
#		for ($curCV=0; $curCV < 10; $curCV++) {
    for ($curCV=0; $curCV < $numCrossVals; $curCV++) {
			$curCVFile = $curCV + 1;

			ReportStdout("Writing CV dataset $curCV");

			open(TRN_CV_FILE, ">${cwd}/${datasetFilename}.${curCVFile}.trn") or 
				ReportFatalError("Cannot open dataset for writing: " . 
												 "${cwd}/${datasetFilename}.${curCVFile}.trn: $!");
			open(TST_CV_FILE, ">${cwd}/${datasetFilename}.${curCVFile}.tst") or 
				ReportFatalError("Cannot open dataset for writing: " . 
												 "${cwd}/${datasetFilename}.${curCVFile}.tst: $!");

			for ($curTrnInd=0; $curTrnInd < $trainCVSize; $curTrnInd++) {
				print TRN_CV_FILE 
					"$alternatedArray[$trainingCVIndices[$curCV][$curTrnInd]]\n";
			}
			for ($curTstInd=0; $curTstInd < $testCVSize; $curTstInd++) {
				print TST_CV_FILE 
					"$alternatedArray[$testingCVIndices[$curCV][$curTstInd]]\n";
			}

			close(TRN_CV_FILE);
			close(TST_CV_FILE);
		}
	}

}

##############################################################################
# Subroutine: TrainAndTest - Train and test using method
#
# Accepts: Hash of configuration info
# Returns: Nothing
##############################################################################
sub TrainAndTest
{
	my $configHashRef = shift;
	my $curCV = shift;
	if($configHashRef->{methodType} eq 'GPNN') {
		return TrainAndTestGPNN($configHashRef, $curCV);
	}
  elsif($configHashRef->{methodType} eq 'GP') {
    return TrainAndTestGP($configHashRef, $curCV);
  }
  elsif($configHashRef->{methodType} eq 'GENN'){
    return TrainAndTestGENN($configHashRef, $curCV);
  }
	elsif($configHashRef->{methodType} eq 'NICO') {
		return TrainAndTestNICO($configHashRef, $curCV);
	}
	else {
		ReportFatalError("TrainAndTest: " . 
										 $configHash{methodType} .
										 " Not Implemented");
	}
}

##############################################################################
# Subroutine: TrainAndTestMPI - Train and test using method
#
# Accepts: Hash of configuration info
# Returns: Nothing
##############################################################################
sub TrainAndTestMPI
{
	my $configHashRef = shift;
	my $curCV = shift;
	if($configHashRef->{methodType} eq 'GPNN') {
			die "no train or test for gpnn-mpi\n";
		return TrainAndTestGPNN($configHashRef, $curCV);
	}
  elsif($configHashRef->{methodType} eq 'GP') {
			die "no train or test for gp-mpi\n";
    return TrainAndTestGP($configHashRef, $curCV);
  }
  elsif($configHashRef->{methodType} eq 'GENN'){
    return TrainAndTestGENNMPI($configHashRef, $curCV);
  }
	elsif($configHashRef->{methodType} eq 'NICO') {
			die "no train or test for nico-mpi\n";
		return TrainAndTestNICO($configHashRef, $curCV);
	}
	else {
		ReportFatalError("TrainAndTest: " . 
										 $configHash{methodType} .
										 " Not Implemented");
	}
}




##############################################################################
# Subroutine: Cleanup - Clean up any temporary files
#
# Accepts: Hash of configuration info
# Returns: Nothing
##############################################################################
sub Cleanup
{
	my $configHashRef = shift;

	if($configHashRef->{methodType} eq 'GPNN') {
		return CleanupGPNN($configHashRef);
	}
  elsif($configHashRef->{methodType} eq 'GP') {
    return CleanupGP($configHashRef);
  }
  elsif($configHashRef->{methodType} eq 'GENN') {
    return CleanupGENN($configHashRef);
  }
	elsif($configHashRef->{methodType} eq 'NICO') {
		return CleanupNICO($configHashRef);
	}
	else {
		ReportFatalError("Cleanup: " . 
										 $configHash{methodType} .
										 " Not Implemented");
	}
}

##############################################################################
# Subroutine: FisherYatesShuffle - generate a random permutation of passed 
#                                  array
#
# Accepts: Array to shuffle
# Returns: Nothing
##############################################################################
sub FisherYatesShuffle {
	my $array = shift;
	my $i;
	for ($i = @$array - 1; $i > 0; $i--) {
		my $j = int rand ($i + 1);
		next if $i == $j;
		@$array[$i, $j] = @$array[$j, $i];
	}
}

##############################################################################
# Subroutine: SystemCall - perform a system call with full error reporting
#
# Accepts: Command to run
# Returns: Success or failure (1/0)
##############################################################################
sub SystemCall
{
	my $commandToRun = shift;

	# assume failure!
	my $ok = 0;

	# run the system command and report the exit status
  my $systemReturnCode = system($commandToRun);
	my $stdoutMsg = sprintf("system() call returned %#04x: ", 
													$commandToRun,
													$systemReturnCode);
	# ReportStdout($stdoutMsg);

	# report a nice message about the system call
	$stdoutMsg = "$commandToRun ";
	if($systemReturnCode == 0) {
		#		$stdoutMsg .= "ran with normal exit";
		$ok = 1;
	}
	elsif($systemReturnCode == 0xff00) {
		$stdoutMsg .= "failed: $!";
		ReportStdout($stdoutMsg);
	}
	elsif(($systemReturnCode & 0xff) == 0) {
		$systemReturnCode >>= 8;
		$stdoutMsg .= "ran with non-zero exit status $systemReturnCode";
		ReportStdout($stdoutMsg);
	}
	else {
		$stdoutMsg .= "ran with ";
		if($systemReturnCode & 0x80) {
			$systemReturnCode &= ~0x80;
			$stdoutMsg .= "coredump from ";
		}
		$stdoutMsg .= "signal $systemReturnCode";
		ReportStdout($stdoutMsg);
	}

	return $ok;
}

##############################################################################
# Subroutine:  GetDatasetCharacteristics
#
# Accepts: Data file name
# Returns: Number of lines and variables per line in the file
##############################################################################
sub GetDatasetCharacteristics
{
	my $datasetFilename = shift;

	# get information about the data file from the file itself
	my ($firstLine, @parsedCols, $cols, $numLines);
	open(THIS_FILE, "<$datasetFilename") or 
		ReportFatalError("Cannot open training data file, $datasetFilename: $!");
	$firstLine = <THIS_FILE>;
	$rows = 1;
	@parsedCols = split /\s+/, $firstLine;
	$cols = scalar(@parsedCols);
	while(<THIS_FILE>) {
		$rows++;
	}
	close(THIS_FILE);

	return ($rows, $cols);
}

##############################################################################
# Subroutine:  AskUser - prompt user and return response
#
# Accepts: Message to prompt user
# Returns: User response
##############################################################################
sub AskUser
{
  my $prompt = shift; 
  my $response;

  print $prompt;
  $response = <STDIN>;

  chomp($response);

  return ($response);
}

##############################################################################
# Subroutine:  UnbalancedDataSplit -  Huiming's function to split data for 10
# fold cross validation, when the data set is unbalanced
#
# Accepts: filename, numCases, numControls
# Returns: a list of teo arrays containing training and testing individual counts
##############################################################################
sub UnbalancedDataSplit
{

  $configHashRef = shift;
  my $numCrossVals = $configHashRef->{crossValidation} || 10;
  
	$filename = shift;
	$groupOne= shift;
	$groupTwo = shift;
        my $cwd = $configHashRef->{workingDirectory};
        my $datasetFilename  = $configHashRef->{baseDatasetName};

	@trainCounts = ();
	@testCounts = ();
	
	open(IF,"<$filename") or die "Can't open file: $filename\n";
	@file = <IF>;
	close IF;

	# print STDOUT "$filename\n";

#	$eachcv1 = int($groupOne/10);
  $eachcv1 = int($groupOne/$numCrossVals);
#	$left1 = $groupOne%10;
	$left1 = $groupOne%$numCrossVals;
	# print("g1 $groupOne eachcv1 $eachcv1 left1 $left1\n");

#	$eachcv2 = int($groupTwo/10);
	$eachcv2 = int($groupTwo/$numCrossVals);
#	$left2 = $groupTwo%10;
	$left2 = $groupTwo%$numCrossVals;
	# print("g2 $groupTwo eachcv2 $eachcv2 left2 $left2\n");


	# print STDOUT "$groupOne  $groupTwo\n";
	(undef,$lines,undef,undef,undef) = split /\s+/, `wc $filename`;
	# print STDOUT "$lines\n";
#	$tstObs = int($lines/10);
	$tstObs = int($lines/$numCrossVals);
	# print STDOUT "$tstObs\n";
	#$filename =~ s/\.txt//;

	if ($lines!=($groupOne+$groupTwo)) {
		print("The number of lines is not correct.\n");
		exit;
	}

#	for ($m = 1; $m <= 10; $m++) {
	for ($m = 1; $m <= $numCrossVals; $m++) {
#print "opened training file $filename.$m.trn\n";exit;
print  "$cwd\n";
print  "$datasetFilename\n";
#		open(TRAIN, ">$filename.$m.trn") or die "Can't open \"$filename.trn\"\n";
open(TRAIN, ">$cwd/$datasetFilename.$m.trn") or die "Can't open \"{$cwd}/$datasetFilename.$m.trn\"\n";

	#	open(TEST, ">$filename.$m.tst") or die "Can't open \"$filename.txt\"\n";
open(TEST, ">$cwd/$datasetFilename.$m.tst") or die "Can't open \"{$cwd}/$datasetFilename.$m.tst\"\n";
		$lineCount = 1;

		#determine the two edges for two groups, which are for test data(1/10).
	  #groupOne split
		if ($m<=$left1) {
			$lowedge1 = ($m-1)*$eachcv1+$m;
			$upedge1 = $m*$eachcv1+$m;
	  } else {
			$lowedge1 = ($m-1)*$eachcv1+$left1+1;
			$upedge1 = $m*$eachcv1+$left1;
	  }

	  #groupTwo split
		if ($m<=$left2) {
			$lowedge2 = ($m-1)*$eachcv2+$m;
			$upedge2 = $m*$eachcv2+$m;
	  } else {
			$lowedge2 = ($m-1)*$eachcv2+$left2+1;
			$upedge2 = $m*$eachcv2+$left2;
	  }
	  $lowedge2 += $groupOne;
	  $upedge2 += $groupOne;
		# print("$m\nl1 $lowedge1 u1 $upedge1 \n");
		# print("l2 $lowedge2 u2 $upedge2 \n\n");

		$trainCount = 0;
		$testCount = 0;
		foreach $k (@file) {
			chomp $k;
		
			#groupOne
			if ($groupOne>=$lineCount) {
				if ($lineCount <= $upedge1 && $lineCount >= $lowedge1) {
					print TEST "$k\n" ;
					$testCount++;
				} else {
					print TRAIN "$k\n" ;
					$trainCount++;
				}
			} else {
				#groupTwo
				if ($lineCount <= $upedge2 && $lineCount >=$lowedge2 ) {
					print TEST "$k\n";
					$testCount++;
				} else {
					print TRAIN "$k\n";
					$trainCount++;
				}
			}

			$lineCount++;
		}

		ReportStdout("Split $m: train: $trainCount, test: $testCount");
		$trainCounts[$m] = $trainCount;
		$testCounts[$m] = $testCount;

		close TRAIN;
		close TEST;
	}

	return;
}


# signal successful module load
ReportStdout("Successfully loaded CrossValidation module");
1;
