#!/usr/bin/perl
# GENN CrossValidation.pm - Scott Dudek - 7/27/04
#
# Modified version of
# GPNNCrossValidation.pm - Bill White - 2/24/02



##############################################################################
# Subroutine: WriteGENNGrammarFile - Write a GENN grammar file
#                                        for a cross validation number
# Accepts: hash of configuration info
# Returns: Nothing
##############################################################################
sub WriteGENNGrammarFile{
  my $configHashRef = shift;

  # this is template grammar file to use
  my $gramfile = $configHashRef->{methodGrammarFile};
  my $wRepeats = $configHashRef->{maxWeightInput} || 2;
  my $leftDigits;# = $configHashRef->{constantLeftDigits} || 2;
  if(defined($configHashRef->{constantLeftDigits})){
    $leftDigits = $configHashRef->{constantLeftDigits};
  }
  else{
    $leftDigits = 2;
  }
  my $rightDigits;# = $configHashRef->{constantRightDigits} || 2;
  if(defined($configHashRef->{constantRightDigits})){
    $rightDigits = $configHashRef->{constantRightDigits};
  }
  else{
    $rightDigits = 2;
  }
  my $zeroDigits = $configHashRef->{zeroDigits} || 0;
  my $numVariables = $configHashRef->{datasetCols}-1;
  
  my $bdn = $configHashRef->{baseDatasetName};
  my $cwd = $configHashRef->{workingDirectory};
  
  my $newgramfile = "${cwd}/${bdn}.genn.gram";
  
  open(GRAM, $gramfile) or 
    ReportFatalError("Cannot open config template: $gramfile: $!");
  my @lines = <GRAM>;
  close(GRAM);
  
  open(NEWGRAM, ">$newgramfile") or
    ReportFatalError("Cannot open config template: $newgramfile: $!");
  
  for(my $i=0; $i<@lines; $i++){
    # check for terminals list 
    if($lines[$i] =~ /^\s*T\s*=/){ 
      my $terminalLine = $lines[$i];
      # if terminals goes past one line concatenate each line
      while($terminalLine !~ /}/){
        $terminalLine .= $lines[++$i];
      }
      # take out spaces
      $terminalLine =~ s/\s//g;
      my $terminalList = ($terminalLine =~ /{(.+)}/)[0];

      # split on commas
      my @terminals = split(/,/, $terminalList);
      # select any that aren't V# like V1, V2, etc.
      my @newterminals;
      foreach my $term(@terminals){
        unless($term =~ /V\d+/){
          push(@newterminals, $term);
        }
      }
      
      # now add list of variables to terminals
      push(@newterminals, "V1-$numVariables");
#      for(my $v=1; $v<=$numVariables; $v++){
#        push(@newterminals, "V$v");
#      }
      
      $terminalLine = join(',', @newterminals);
      $terminalLine = 'T = {' . $terminalLine . "}\n";
      print NEWGRAM $terminalLine;
    }
    elsif($lines[$i] =~ /<pinput>\s*::=/){
      # first is w=2 always so print out
      print NEWGRAM $lines[$i];
      
      # create as many possible weighted inputs as needed
      for(my $w=3; $w<=$wRepeats; $w++){
        print NEWGRAM "            | ";
        for(my $wcount=1; $wcount<=$w; $wcount++){
          print NEWGRAM 'W(<winput>), ';
        }
        print NEWGRAM "$w\n";
      }
      print NEWGRAM "\n";
      # skip lines until next rule starts
      while($lines[++$i] =~ /\|/){}
    }
    elsif($lines[$i] =~ /<num>\s*::=/){
      my ($digits, $totaldigits);
      for(my $i=1; $i<=$leftDigits; $i++){
        $digits .= "<dig> ";
        $totaldigits++;
      }
      $digits .= ". ";
      $totaldigits++;
      # add any constant zeroes to the possible weight
      for(my $i=1; $i<=$zeroDigits; $i++){
        $digits .= "0 ";
        $totaldigits++;
      }
      for(my $i=1; $i<=$rightDigits; $i++){
        $digits .= "<dig> ";
        $totaldigits++;
      }
      print NEWGRAM "<num>     ::= $digits $totaldigits\n\n";
       # skip lines until next rule starts
      while($lines[++$i] =~ /\|/){}
    }
    elsif($lines[$i] =~ /<v>\s*::=/){
      print NEWGRAM "<v>       ::= V1-$numVariables\n";
#      for($v=2; $v<=$numVariables; $v++){
#        print NEWGRAM "            | V$v\n";
#      }
#      print NEWGRAM "\n";
      while($lines[++$i] =~ /\|/){}
    }
    else{
      print NEWGRAM $lines[$i];
    } 
  }
  
  close(NEWGRAM);
}



##############################################################################
# Subroutine: WriteGENNConfigFile - Write a GENN training config file
#                                        for a cross validation number
# Accepts: Cross validation number, hash of configuration info
# Returns: Nothing
##############################################################################
sub WriteGENNConfigFile
{
	my $cvNum = shift;
	my $configHashRef = shift;

	my $templateFilename = $configHashRef->{methodConfigTemplate};

	my $bdn = $configHashRef->{baseDatasetName};
	my $bdd = $configHashRef->{baseDatasetDirectory};

	my $cwd = $configHashRef->{workingDirectory};
	my $exd = $configHashRef->{exeDirectory};

	my $newFilename = "${cwd}/${bdn}.${cvNum}.genn.config";
	my $newTrainDatasetFilename = "${cwd}/${bdn}.${cvNum}.trn";
	my $newTestDatasetFilename = "${cwd}/${bdn}.${cvNum}.tst";
		
	my $newGrammarFilename = "${cwd}/${bdn}.genn.gram";
	my $numVariables = $configHashRef->{datasetCols}-1; 
        my $codonsize;
        if($numVariables <= 256){
	  $codonsize = 8;
	}
	else{
	  $codonsize = 16;
	}
	
	open(TEMPLATE_FILE,"<$templateFilename") or 
		ReportFatalError("Cannot open config template: $templateFilename: $!");
  open(NEW_CONFIG_FILE,">$newFilename") or
		ReportFatalError("Cannot open new config file: $newFilename: $!");
	my ($paramName, $paramValue);
  while(<TEMPLATE_FILE>) {
		chomp;
    ($paramName, $paramValue) = split /=/;
		$paramName =~ s/\s+//g;
		$paramValue =~ s/\s+//g;
    if($paramName eq 'expression_filename') {
      printf(NEW_CONFIG_FILE "%s = %s\n", 
						 $paramName, $newDatasetFilename);
#    } elsif($paramName eq 'number_of_individuals') {
#      printf(NEW_CONFIG_FILE "%s = %d\n", 
#						 $paramName, $trainCounts[$cvNum]);
#    } elsif($paramName eq 'number_of_genes') {
#      printf(NEW_CONFIG_FILE "%s = %s\n", 
#						 $paramName, $configHashRef->{datasetCols}-1);
    } elsif($paramName eq 'random_seed') {
      
#      printf(NEW_CONFIG_FILE "%s = %s\n", 
#						 $paramName, $configHashRef->{randomSeed});
      my $randomSeedNum = $configHashRef->{randomSeed} + $cvNum * $configHashRef->{gennNumProcesses};
			printf(NEW_CONFIG_FILE "%s = %s\n", 
						 $paramName, $randomSeedNum);
    } elsif($paramName eq 'output.basename') {
      printf(NEW_CONFIG_FILE "%s = %s\n", 
						 $paramName, "${cwd}/${bdn}.${cvNum}");
	  } elsif($paramName eq 'train_file') {
      printf(NEW_CONFIG_FILE "%s = %s\n", 
						 $paramName, $newTrainDatasetFilename);
	  } elsif($paramName eq 'test_file') {
      printf(NEW_CONFIG_FILE "%s = %s\n", 
						 $paramName, $newTestDatasetFilename);
	  } elsif($paramName eq 'grammar_file') {
      printf(NEW_CONFIG_FILE "%s = %s\n", 
						 $paramName, $newGrammarFilename);
          } elsif($paramName eq 'codon_size'){
           printf(NEW_CONFIG_FILE "%s = %s\n", $paramName, $codonsize);
 
    } else {
      printf(NEW_CONFIG_FILE "%s\n", $_);
    }
  }
  close(NEW_CONFIG_FILE);
  close(TEMPLATE_FILE);
}

##############################################################################
# Subroutine: TrainAndTestGENN - Train and test using GENN
#
# Accepts: Hash of configuration info, current CV
# Returns: Nothing
##############################################################################
sub TrainAndTestGENN
{

	my $configHashRef = shift;
	my $curCV = shift;

	my $cwd = $configHashRef->{workingDirectory};
	my $exd = $configHashRef->{exeDirectory};

	my $bdn = $configHashRef->{baseDatasetName};
	my $bdd = $configHashRef->{baseDatasetDirectory};
	
	my $optimizeMisclassification = -1;
	

	my $trnDatasetFilename = "${cwd}/${bdn}.$curCV.trn";
	my $tstDatasetFilename = "${cwd}/${bdn}.$curCV.tst";

  my $outputFile = "${cwd}/${bdn}.${curCV}.out";

  my $gramFile = "${cwd}/${bdn}.genn.gram";
  my $configGENNFile = "${cwd}/${bdn}.${curCV}.genn.config";

	# FIX ME: move these to config file?
	my $numProcessesToSpawn = $configHashRef->{gennNumProcesses};
	my $pvmHostsFilename    = $configHashRef->{gennHostsFile};

	if(!defined($configHashRef->{gennNumProcesses})) {
		ReportFatalError("TrainAndTestGENN: \"gennNumProcesses\" must be " . 
										 "defined in the config file");
	}

	if(!defined($configHashRef->{gennHostsFile})) {
		ReportFatalError("TrainAndTestGENN: \"gennHostsFile\" must be " . 
										 "defined in the config file");
	}

	# --------------------------------------------------------------------------
	# training and testing are done sequentially using same
	# call to GENN
	# create CV model using training data
	#  ReportStdout("GENN: CV-${curCV}");

  #my $command = "${exd}/genn $configGENNFile";
  # pvm version
	my $command = "${exd}/gennMaster $configGENNFile $numProcessesToSpawn ${exd}/gennSlave";
ReportStdout($command);
	my $ok = SystemCall($command);
	if(!$ok){
	  ReportFatalError("TrainAndTestGENN: Error running $command " . 
										 "for training: $ok");
	}
	# wait until training has completed and written best model file
#  GENNWaitUntilFileExists("${trnDatasetFilename}.individual.bst");
	
	# print misclassification error
  my $grepCommand = "grep \"training error\" $outputFile";
	my $grepResults = `$grepCommand`;
  my ($Junk, $trainingMisclassification) = split(/=/, $grepResults);
	chomp($trainingMisclassification);
  my $stdoutMsg = sprintf("GENN: %2d: Training misclassification: %6.4f", 
													$curCV, $trainingMisclassification);
	ReportStdout($stdoutMsg);
	$grepCommand = "grep \"testing error\" $outputFile";
	my $grepResults = `$grepCommand`;
	my ($Junk, $testingMisclassification) = split(/=/, $grepResults);
	$stdoutMsg = sprintf("GENN: %2d: Testing misclassification: %6.4f", 
													$curCV, $testingMisclassification);
	ReportStdout($stdoutMsg);


	# --------------------------------------------------------------------------
  # create list of loci in the current model
  my @locusList = GENNLociInModel($outputFile);
	# save loci used in each model for reporting
	my $modelLoci = join(" ", @locusList);
  $stdoutMsg = sprintf("GENN: %2d: Loci used in the model: %s", 
											 $curCV, $modelLoci);
	ReportStdout($stdoutMsg);

	my %resultsHash = (
										 lociUsed => $modelLoci,
										 trnMisclass => $trainingMisclassification,
										 tstMisclass => $testingMisclassification,
										 optMisclass => $optimizeMisclassification
										);
	
	return %resultsHash;
}


##############################################################################
# Subroutine: TrainAndTestGENN - Train and test using GENN
#
# Accepts: Hash of configuration info, current CV
# Returns: Nothing
##############################################################################
sub TrainAndTestGENNMPI
{

		print "Starting TrainAndTestGENNMPI\n";

	my $configHashRef = shift;
	my $curCV = shift;

	my $cwd = $configHashRef->{workingDirectory};
	my $exd = $configHashRef->{exeDirectory};

	my $bdn = $configHashRef->{baseDatasetName};
	my $bdd = $configHashRef->{baseDatasetDirectory};
	
	my $optimizeMisclassification = -1;
	

	my $trnDatasetFilename = "${cwd}/${bdn}.$curCV.trn";
	my $tstDatasetFilename = "${cwd}/${bdn}.$curCV.tst";

  my $outputFile = "${cwd}/${bdn}.${curCV}.out";

  my $gramFile = "${cwd}/${bdn}.genn.gram";
  my $configGENNFile = "${cwd}/${bdn}.${curCV}.genn.config";

	# FIX ME: move these to config file?
	my $numProcessesToSpawn = $configHashRef->{gennNumProcesses};
	my $pvmHostsFilename    = $configHashRef->{gennHostsFile};

	if(!defined($configHashRef->{gennNumProcesses})) {
		ReportFatalError("TrainAndTestGENN: \"gennNumProcesses\" must be " . 
										 "defined in the config file");
	}

	if(!defined($configHashRef->{gennHostsFile})) {
		ReportFatalError("TrainAndTestGENN: \"gennHostsFile\" must be " . 
										 "defined in the config file");
	}

	# --------------------------------------------------------------------------
	# training and testing are done sequentially using same
	# call to GENN
	# create CV model using training data
	#  ReportStdout("GENN: CV-${curCV}");

  #my $command = "${exd}/genn $configGENNFile";
  # pvm version
	#my $command = "${exd}/gennMaster $configGENNFile $numProcessesToSpawn ${exd}/gennSlave";

		#my $thisSeed = $configHashRef->{randomSeed};
		#print "Seed is $thisSeed\n";
		#my $schemaname = "${cwd}/appschema";
		#print "Creating schema file for $numProcessesToSpawn slaves to $schemaname\n";
	#open(SCHEMAOUT, ">$schemaname") or die "Unable to open appschema file";

#	print SCHEMAOUT "c0 ${exd}/gennMaster $configGENNFile $numProcessesToSpawn ${exd}/gennSlave\n";

#	for (my $i = 0; $i < $numProcessesToSpawn; $i++) {
	#		my $randseed = $thisSeed + $i;
		#	print SCHEMAOUT "c0 ${exd}/gennSlave $configGENNFile $randseed\n";
	#}

	#close(SCHEMAOUT);

		#print "Schema file written\n";

		#my $command = "/usr/local/bin/mpirun $schemaname";

		$numMPIProcesses = $numProcessesToSpawn + 1;
		my $command = "mpiexec -n $numMPIProcesses ${exd}/gennMPI $configGENNFile $numProcessesToSpawn /dev/null";

ReportStdout($command);
	my $ok = SystemCall($command);
	if(!$ok){
	  ReportFatalError("TrainAndTestGENN: Error running $command " . 
										 "for training: $ok");
	}
	# wait until training has completed and written best model file
#  GENNWaitUntilFileExists("${trnDatasetFilename}.individual.bst");
	
	# print misclassification error
  my $grepCommand = "grep \"training error\" $outputFile";
	my $grepResults = `$grepCommand`;
  my ($Junk, $trainingMisclassification) = split(/=/, $grepResults);
	chomp($trainingMisclassification);
  my $stdoutMsg = sprintf("GENN: %2d: Training misclassification: %6.4f", 
													$curCV, $trainingMisclassification);
	ReportStdout($stdoutMsg);
	$grepCommand = "grep \"testing error\" $outputFile";
	my $grepResults = `$grepCommand`;
	my ($Junk, $testingMisclassification) = split(/=/, $grepResults);
	$stdoutMsg = sprintf("GENN: %2d: Testing misclassification: %6.4f", 
													$curCV, $testingMisclassification);
	ReportStdout($stdoutMsg);


	# --------------------------------------------------------------------------
  # create list of loci in the current model
  my @locusList = GENNLociInModel($outputFile);
	# save loci used in each model for reporting
	my $modelLoci = join(" ", @locusList);
  $stdoutMsg = sprintf("GENN: %2d: Loci used in the model: %s", 
											 $curCV, $modelLoci);
	ReportStdout($stdoutMsg);

	my %resultsHash = (
										 lociUsed => $modelLoci,
										 trnMisclass => $trainingMisclassification,
										 tstMisclass => $testingMisclassification,
										 optMisclass => $optimizeMisclassification
										);
	
	return %resultsHash;
}

##############################################################################
# Subroutine: CleanupGENN - clean up all temp files used during CV
#
# Accepts: Hash of configuration info
# Returns: Nothing
##############################################################################
sub CleanupGENN
{
	my $configHashRef = shift;
	my $curCV;
	my $baseName = $configHashRef->{baseDatasetName};
	my $cwd      = $configHashRef->{workingDirectory};
  my $numCrossVals = $configHashRef->{crossValidation} || 10;
  
  unlink "${cwd}/${baseName}.genn.gram";

#	for($curCV = 1; $curCV <= 10; $curCV++) {
  for($curCV = 1; $curCV <= $numCrossVals; $curCV++) {
		SystemCall("mv ${baseName}.${curCV}.out smd");
		#SystemCall("mkdir temp");
		#SystemCall("mv *.debug temp");
 		my @debugfiles = glob("$baseName.${curCV}.*.debug");
                if(@debugfiles){
			for(my $curDebug=1; $curDebug<=@debugfiles; $curDebug++){
			  SystemCall("mv $debugfiles[$curDebug-1] tmp.debug.$curDebug");
                        } 
		}
		unlink glob("$baseName.${curCV}.*");
		SystemCall("mv smd $baseName.$curCV.model");
		if(@debugfiles){
                        for(my $curDebug=1; $curDebug<=@debugfiles; $curDebug++){
                          SystemCall("mv tmp.debug.$curDebug $debugfiles[$curDebug-1]");
                        }
                }
		unlink "${baseName}.${curCV}.genn.config";
		#SystemCall("mv temp/* .");
		#SystemCall("rmdir temp");
		#SystemCall("rm run.${curCV}.log");
	}
	unlink "sa_fitness.temp";
}

###################### F R O M   C R O S S _ V A L I D A T E 2 ################

###############################
#
# Subroutine:  GENNCalculateMisclass
#
# Input: $TestingValuesFile, $TestingDataFile
# Returns: Testing misclassification
sub GENNCalculateMisclass
{
	# get passed params
  my ($TestingValuesFile, $TestingDataFile) = @_;
	my (@testingStatuses, @classificationCounts);
	my ($lineCount, @lineValues);
	my $misclassError = 1.0;

	# read in the testing data
	$lineCount = 0;
	open(DATA_FILE, "<$TestingDataFile") or 
		die("Cannot open testing data file $TestingDataFile: $!");
	while(<DATA_FILE>) {
		@lineValues = split /\s+/;
		$testingStatuses[$lineCount++] = $lineValues[scalar(@lineValues) - 1];
	}
	close(DATA_FILE);

	# read in values from evaluations, compare to statuses and accumulate counts
	$lineCount = 0;
	open(VALUES_FILE, "<$TestingValuesFile") or 
		die("Cannot open testing values file $TestingValuesFile: $!");
	# put classification counts in matrix classificationCounts
	while(<VALUES_FILE>) {
		$classificationCounts[$testingStatuses[$lineCount++]][$_]++;
	}
	close(VALUES_FILE);

	$misclassError = 
		($classificationCounts[0][1] + $classificationCounts[1][0]) / 
			$lineCount;

	return $misclassError;
}

###############################
#
# Subroutine:  GENNWaitUntilFileExists
#
# Wait for file to exist or until PVMSettleTime seconds
#
sub GENNWaitUntilFileExists
{
  my ($FileName) = @_;

	# FIXME: add to config file?
	$PVMSettleTime = 10;

  $StartDate = `date`;
  ($Junk1, $Junk2, $Junk3, $StartDate, $RestJunk) = split(/ +/, $StartDate);
  ($Junk1, $Junk2, $StartDate) = split(/:/, $StartDate);

  $DiffTime = 0;
	#	print "Checking for existence of $FileName\n";
  $sysReturn = SystemCall("ls $FileName > .trash.cv.FileExists");
  while((!$sysReturn) && ($DiffTime < $PVMSettleTime)) {
    $sysReturn = SystemCall("ls $FileName >& .trash.cv.FileExists");
    $NewDate = `date`;
    ($Junk1, $Junk2, $Junk3, $NewDate, $RestJunk) = split(/ +/, $NewDate);
    ($Junk1, $Junk2, $NewDate) = split(/:/, $NewDate);
    if($NewDate < $StartDate) {
      $StartDate = $StartDate - 60;
    }
    $DiffTime = $NewDate - $StartDate;
  }
  if(!$sysReturn) {
    ReportFatalError("After waiting $PVMSettleTime seconds, $FileName has " .
										 "not yet arrived! Aborting...");
  }
  SystemCall("rm -f .trash.cv.FileExists");

  return($sysReturn);
}

###############################
#
# Subroutine:  GENNLociInModel
#
#   Input:  File Name
# Returns:  A list of the locus numbers
# Purpose:  Create a list of function loci included in the model function.
sub GENNLociInModel
{
  my ($FileName) = @_;
  my (@LocusList);

  open(IN, $FileName) or 
    ReportFatalError("Cannot open genn output file: $FileName: $!");
  
  while(<IN>){
    push(@LocusList, (/V(\d+)/g));
  }
  
  close(IN);
  
  @LocusList = sort {eval($a) <=> eval($b) } @LocusList;
  
  return (@LocusList);
  
  
} # end of LociInModel

# signal successful module load
ReportStdout("Successfully loaded GENNCrossValidation module");
1;
