#!/usr/bin/perl
# GP CrossValidation.pm - Scott Dudek - 7/27/04
#
# Modified version of
# GPNNCrossValidation.pm - Bill White - 2/24/02

##############################################################################
# Subroutine: WriteGPNNTrainConfigFile - Write a GPNN training config file
#                                        for a cross validation number
# Accepts: Cross validation number, hash of configuration info
# Returns: Nothing
##############################################################################
sub WriteGPTrainConfigFile
{
	my $cvNum = shift;
	my $configHashRef = shift;

	my $templateFilename = $configHashRef->{methodConfigTemplate};

	my $bdn = $configHashRef->{baseDatasetName};
	my $bdd = $configHashRef->{baseDatasetDirectory};

	my $cwd = $configHashRef->{workingDirectory};
	my $exd = $configHashRef->{exeDirectory};

	my $newFilename = "${cwd}/${bdn}.${cvNum}.trn.config";
	my $newDatasetFilename = "${cwd}/${bdn}.${cvNum}.trn";

	open(TEMPLATE_FILE,"<$templateFilename") or 
		ReportFatalError("Cannot open config template: $templateFilename: $!");
  open(NEW_CONFIG_FILE,">$newFilename") or
		ReportFatalError("Cannot open new config file: $newFilename: $!");
	my ($paramName, $paramValue);
  while(<TEMPLATE_FILE>) {
		chomp;
    ($paramName, $paramValue) = split /=/;
		$paramName =~ s/\s+//g;
		$paramValue =~ s/\s+//g;
    if($paramName eq 'expression_filename') {
      printf(NEW_CONFIG_FILE "%s = %s\n", 
						 $paramName, $newDatasetFilename);
    } elsif($paramName eq 'number_of_individuals') {
      printf(NEW_CONFIG_FILE "%s = %d\n", 
						 $paramName, $trainCounts[$cvNum]);
    } elsif($paramName eq 'number_of_genes') {
      printf(NEW_CONFIG_FILE "%s = %s\n", 
						 $paramName, $configHashRef->{datasetCols}-1);
    } elsif($paramName eq 'random_seed') {
      my $randomSeedNum = $configHashRef->{randomSeed} + $cvNum * $configHashRef->{gpNumProcesses};
			printf(NEW_CONFIG_FILE "%s = %s\n", 
						 $paramName, $randomSeedNum);
#      printf(NEW_CONFIG_FILE "%s = %s\n", 
#						 $paramName, $configHashRef->{randomSeed});
    } elsif($paramName eq 'output.basename') {
      printf(NEW_CONFIG_FILE "%s = %s\n", 
						 $paramName, $newDatasetFilename);
    } else {
      printf(NEW_CONFIG_FILE "%s\n", $_);
    }
  }
  close(NEW_CONFIG_FILE);
  close(TEMPLATE_FILE);
}

##############################################################################
# Subroutine: WriteGPTestConfigFile - Write a GP testing config file
#                                       for a cross validation number
# Accepts: Cross validation number, hash of configuration info
# Returns: Nothing
##############################################################################
sub WriteGPTestConfigFile
{
	my $cvNum = shift;
	my $configHashRef = shift;
	
	my $templateFilename = $configHashRef->{methodConfigTemplate};

	my $bdn = $configHashRef->{baseDatasetName};
	my $bdd = $configHashRef->{baseDatasetDirectory};

	my $cwd = $configHashRef->{workingDirectory};
	my $exd = $configHashRef->{exeDirectory};

	my $newFilename = "${cwd}/${bdn}.${cvNum}.tst.config";
	my $newDatasetFilename = "${cwd}/${bdn}.${cvNum}.tst";

  open(TEMPLATE_FILE,"<$templateFilename") or 
		ReportFatalError("Cannot open config template: $templateFilename: $!");
  open(NEW_CONFIG_FILE,">$newFilename") or
		ReportFatalError("Cannot open new config file: $newFilename: $!");
	my ($paramName, $paramValue);
  while(<TEMPLATE_FILE>) {
		chomp;
    ($paramName, $paramValue) = split /=/;
		$paramName =~ s/\s+//g;
		$paramValue =~ s/\s+//g;
    if($paramName eq 'expression_filename') {
      printf(NEW_CONFIG_FILE "%s = %s\n", 
						 $paramName, $newDatasetFilename);
    } elsif($paramName eq 'number_of_individuals') {
      printf(NEW_CONFIG_FILE "%s = %d\n", 
						 $paramName, $testCounts[$cvNum]);
    } elsif($paramName eq 'number_of_genes') {
      printf(NEW_CONFIG_FILE "%s = %s\n", 
						 $paramName, $configHashRef->{datasetCols}-1);
    } elsif($paramName eq 'random_seed') {
    my $randomSeedNum = $configHashRef->{randomSeed} + $cvNum * $configHashRef->{gpNumProcesses};
			printf(NEW_CONFIG_FILE "%s = %s\n", 
						 $paramName, $randomSeedNum);
#      printf(NEW_CONFIG_FILE "%s = %s\n", 
#						 $paramName, $configHashRef->{randomSeed});
    } elsif($paramName eq 'output.basename') {
      printf(NEW_CONFIG_FILE "%s = %s\n", 
						 $paramName, $bdn . ".${cvNum}.trn");
    } else {
      printf(NEW_CONFIG_FILE "%s\n", $_);
    }
  }
  close(NEW_CONFIG_FILE);
  close(TEMPLATE_FILE);
}

##############################################################################
# Subroutine: TrainAndTestGP - Train and test using GP
#
# Accepts: Hash of configuration info, current CV
# Returns: Nothing
##############################################################################
sub TrainAndTestGP
{
	my $configHashRef = shift;
	my $curCV = shift;

	my $cwd = $configHashRef->{workingDirectory};
	my $exd = $configHashRef->{exeDirectory};

	my $bdn = $configHashRef->{baseDatasetName};
	my $bdd = $configHashRef->{baseDatasetDirectory};

	my $trnDatasetFilename = "${cwd}/${bdn}.$curCV.trn";
	my $tstDatasetFilename = "${cwd}/${bdn}.$curCV.tst";

	my $trnConfigFilename = "$trnDatasetFilename.config";
	my $tstConfigFilename = "$tstDatasetFilename.config";

	# FIX ME: move these to config file?
	my $numProcessesToSpawn = $configHashRef->{gpNumProcesses};
	my $pvmHostsFilename    = $configHashRef->{gpHostsFile};

	if(!defined($configHashRef->{gpNumProcesses})) {
		ReportFatalError("TrainAndTestGP: \"gpNumProcesses\" must be " . 
										 "defined in the config file");
	}

	if(!defined($configHashRef->{gpHostsFile})) {
		ReportFatalError("TrainAndTestGP: \"gpHostsFile\" must be " . 
										 "defined in the config file");
	}

	# --------------------------------------------------------------------------
  # create CV model using training data
	#  ReportStdout("GP: CV-${curCV}");
  my $trainingCommand = 
    "${exd}/gpmaster " .
    "${exd}/gp " .
    "$trnConfigFilename " .
    "$numProcessesToSpawn " .
    "$pvmHostsFilename";
	
	my $ok = SystemCall($trainingCommand);
	if(!$ok) {
		ReportFatalError("TrainAndTestGP: Error running $trainingCommand " . 
										 "for training: $ok");
	}

	# wait until training has completed and written best model file
  GPWaitUntilFileExists("${trnDatasetFilename}.individual.bst");
	
	# print misclassification error
  my $grepCommand = "grep \"raw fitness\" $trnDatasetFilename.bst";
	my $grepResults = `$grepCommand`;
  my ($Junk, $trainingMisclassification) = split(/:/, $grepResults);
	chomp($trainingMisclassification);
  my $stdoutMsg = sprintf("GP: %2d: Training misclassification: %6.4f", 
													$curCV, $trainingMisclassification);
	ReportStdout($stdoutMsg);

	# --------------------------------------------------------------------------
	# test CV model using testing data 
	my $optimizeMisclassification = -1;
	if($configHashRef->{localOptimize}) {
		# added local optimization - bcw - 2/7/03

		# create model file for this CV
		my $testingCommand =
			"${exd}/gp_model_test $tstConfigFilename";
		$ok = SystemCall($testingCommand);
		if(!$ok) {
			ReportFatalError("TrainAndTestGP: Error running $testingCommand " . 
											 "for creating .model file: $ok");
		}

		my $saConfigFilename = $configHashRef->{localOptimizeConfigFile};
		if(!$saConfigFilename) {
			ReportFatalError("TrainAndTestGP: Local optimize config file missing");
		}
		# write dataset names to SA config file
		SystemCall("cp $saConfigFilename $saConfigFilename.thisCV");
		open(SA_CONFIG_FILE, ">>$saConfigFilename.thisCV") or 
			die("Cannot open SA config file: $saConfigFilename: $!");
		print SA_CONFIG_FILE "train_dataset_name = $trnDatasetFilename\n";
		print SA_CONFIG_FILE "test_dataset_name = $tstDatasetFilename\n";
		close(SA_CONFIG_FILE);

		my $gpSAProg = "${exd}/gp_local_optimize";
		my $command = "cat ${cwd}/${bdn}.${curCV}.tst.model | " . 
			"$gpSAProg $saConfigFilename.thisCV > run.${curCV}.log";
#		print "$command\n";
		`$command`;
		unlink "$saConfigFilename.thisCV";

		# read errors from SA output file
		open(SA_ERROR_FILE, "${cwd}/sa_fitness.temp") or 
			die("Cannot open SA fitness file: ${cwd}/sa_fitness.temp: $!");

		# initial training error
		my $tempLine = <SA_ERROR_FILE>;
		chomp $tempLine;
		if(abs($tempLine - $trainingMisclassification) > 0.0001) {
			$stdoutMsg = 
				sprintf("WARNING: lil-gp and gp_local_optimize training ". 
				"errors do not match: %f vs. %f", $trainingMisclassification, 
				$tempLine);
			ReportStdout($stdoutMsg);
		}

		# training error after optimize
		$tempLine = <SA_ERROR_FILE>;
		chomp $tempLine;
		$optimizeMisclassification = $tempLine;

		# prediction error
		$tempLine = <SA_ERROR_FILE>;
		chomp $tempLine;
		$testingMisclassification = $tempLine;

		close(SA_ERROR_FILE);

		$stdoutMsg = sprintf("GP: %2d: Testing misclassification (SA): %6.4f", 
												 $curCV, $testingMisclassification);
	}
	else {
		my $testingCommand =
			"${exd}/gp_model_test $tstConfigFilename";
		$ok = SystemCall($testingCommand);
#print "$testingCommand\n";exit;
		if(!$ok) {
			ReportFatalError("TrainAndTestGP: Error running $testingCommand " . 
											 "for testing: $ok");
		}
		
		$testingMisclassification = 
			GPCalculateMisclass("$tstDatasetFilename.values",	$tstDatasetFilename);
		$stdoutMsg = sprintf("GP: %2d: Testing misclassification:  %6.4f", 
												 $curCV, $testingMisclassification);
	}
	ReportStdout($stdoutMsg);

	# --------------------------------------------------------------------------
  # create list of loci in the current model
  my @locusList = GPLociInModel("$tstDatasetFilename.model");
	# save loci used in each model for reporting
	my $modelLoci = join(" ", @locusList);
  $stdoutMsg = sprintf("GP: %2d: Loci used in the model: %s", 
											 $curCV, $modelLoci);
	ReportStdout($stdoutMsg);

	my %resultsHash = (
										 lociUsed => $modelLoci,
										 trnMisclass => $trainingMisclassification,
										 tstMisclass => $testingMisclassification,
										 optMisclass => $optimizeMisclassification
										);
	
	return %resultsHash;
}

##############################################################################
# Subroutine: CleanupGP - clean up all temp files used during CV
#
# Accepts: Hash of configuration info
# Returns: Nothing
##############################################################################
sub CleanupGP
{
	my $configHashRef = shift;
	my $curCV;
	my $baseName = $configHashRef->{baseDatasetName};
	my $cwd      = $configHashRef->{workingDirectory};
  my $numCrossVals = $configHashRef->{crossValidation} || 10;
  
#	for($curCV = 1; $curCV <= 10; $curCV++) {
  for($curCV = 1; $curCV <= $numCrossVals; $curCV++) {
		SystemCall("mv $baseName.$curCV.tst.model smd");
		#SystemCall("mkdir temp");
		#SystemCall("mv *.debug temp");
		unlink glob("$baseName.${curCV}.*");
		SystemCall("mv smd $baseName.$curCV.tst.model");
		unlink "$baseName_${curCV}.trn.config";
		#SystemCall("mv temp/* .");
		#SystemCall("rmdir temp");
		#SystemCall("rm run.${curCV}.log");
	}
	unlink "sa_fitness.temp";
}

###################### F R O M   C R O S S _ V A L I D A T E 2 ################

###############################
#
# Subroutine:  GPCalculateMisclass
#
# Input: $TestingValuesFile, $TestingDataFile
# Returns: Testing misclassification
sub GPCalculateMisclass
{
	# get passed params
  my ($TestingValuesFile, $TestingDataFile) = @_;
	my (@testingStatuses, @classificationCounts);
	my ($lineCount, @lineValues);
	my $misclassError = 1.0;

	# read in the testing data
	$lineCount = 0;
	open(DATA_FILE, "<$TestingDataFile") or 
		die("Cannot open testing data file $TestingDataFile: $!");
	while(<DATA_FILE>) {
		@lineValues = split /\s+/;
		$testingStatuses[$lineCount++] = $lineValues[scalar(@lineValues) - 1];
	}
	close(DATA_FILE);

	# read in values from evaluations, compare to statuses and accumulate counts
	$lineCount = 0;
	open(VALUES_FILE, "<$TestingValuesFile") or 
		die("Cannot open testing values file $TestingValuesFile: $!");
	# put classification counts in matrix classificationCounts
	while(<VALUES_FILE>) {
		$classificationCounts[$testingStatuses[$lineCount++]][$_]++;
	}
	close(VALUES_FILE);

	$misclassError = 
		($classificationCounts[0][1] + $classificationCounts[1][0]) / 
			$lineCount;

	return $misclassError;
}

###############################
#
# Subroutine:  GPWaitUntilFileExists
#
# Wait for file to exist or until PVMSettleTime seconds
#
sub GPWaitUntilFileExists
{
  my ($FileName) = @_;

	# FIXME: add to config file?
	$PVMSettleTime = 10;

  $StartDate = `date`;
  ($Junk1, $Junk2, $Junk3, $StartDate, $RestJunk) = split(/ +/, $StartDate);
  ($Junk1, $Junk2, $StartDate) = split(/:/, $StartDate);

  $DiffTime = 0;
	#	print "Checking for existence of $FileName\n";
  $sysReturn = SystemCall("ls $FileName > .trash.cv.FileExists");
  while((!$sysReturn) && ($DiffTime < $PVMSettleTime)) {
    $sysReturn = SystemCall("ls $FileName >& .trash.cv.FileExists");
    $NewDate = `date`;
    ($Junk1, $Junk2, $Junk3, $NewDate, $RestJunk) = split(/ +/, $NewDate);
    ($Junk1, $Junk2, $NewDate) = split(/:/, $NewDate);
    if($NewDate < $StartDate) {
      $StartDate = $StartDate - 60;
    }
    $DiffTime = $NewDate - $StartDate;
  }
  if(!$sysReturn) {
    ReportFatalError("After waiting $PVMSettleTime seconds, $FileName has " .
										 "not yet arrived! Aborting...");
  }
  SystemCall("rm -f .trash.cv.FileExists");

  return($sysReturn);
}

###############################
#
# Subroutine:  GPLociInModel
#
#   Input:  File Name
# Returns:  A list of the locus numbers
# Purpose:  Create a list of function loci included in the model function.
sub GPLociInModel
{
  my ($FileName) = @_;
  my (@GrepFile, @LocusList, $LocusIndex, $Trash, %Loci);

  @GrepFile = `grep "Var" $FileName`;
  for($LocusIndex = 0; $LocusIndex < scalar(@GrepFile); $LocusIndex++) {
    chomp($GrepFile[$LocusIndex]);
    # escaped '#' metacharacter in split below - bcw 7/27/01
    # ($Trash, $LocusList[$LocusIndex]) =  split(/ \#/, $GrepFile[$LocusIndex]);
    # ($LocusList[$LocusIndex],$Trash)  =  split(/\)/, $LocusList[$LocusIndex]);
    push(@LocusList, ($GrepFile[$LocusIndex] =~ /Var\s\#(\d+)/g)); 
    #$Loci{($GrepFile[$LocusIndex] =~ /Var\s\#(\d+)/g)[0]} = 1;
  }

  @LocusList = sort { eval($a) <=> eval($b) } @LocusList;
  #@LocusList = sort { eval($a) <=> eval($b) } keys %Loci;
  
  
  return (@LocusList);
  
} # end of LociInModel

# signal successful module load
ReportStdout("Successfully loaded GPCrossValidation module");
1;
