/* ----------------------------------------------------------------------------
  genome.cpp
  mbwall 5dec95
  Copyright (c) 1995-1996 Massachusetts Institute of Technology

 DESCRIPTION:
   This file contains the information needed to evaluate an operate on a 
genome in these examples.  We define routines for sending and receiving genome
data as well as routines for modifying the genetic algorithm's population.
   At some point we should add better error checking to the pvm status returns.
---------------------------------------------------------------------------- */

#include <unistd.h>
#include "genome.h"

// When this flag is defined, the initialize and evaluate function dump a 
// string to cerr that tells which host they're running on.  You can then look
// at this info in the pvm log file.
//#define DEBUG


int BADONES=0;
int GOODONES=0;

///
/// Output to cout the number of valid and invalid (bad) 
/// neural networks counted
///
void printGoodBad(){
  cout << "good=" << GOODONES << " bad=" << BADONES << endl;
  GOODONES = BADONES = 0;
}


///
/// Evaluator for the standard crossover 
///
///
float
GenomeEvaluator(GAGenome& g) {
   GEBinStrBlockCrossGenome& genome = (GEBinStrBlockCrossGenome &) g;
  float total=0.0;
  // convert genome into grammar rules
  trans.convert_codons(genome);
  vector<Element *> term_list;
  vector<Element *> postfix_stack;
  vector<Element *> constants_list; // used to track dynamically created constants
  if(trans.create_terminals(term_list, constants_list)){
    // convert to postfix stack
    postfix_stack=trans.create_postfix_stack(term_list);
    // need to test the genome
    total = 1 - genn_test.determine_error_rate(postfix_stack);
  }

  for(unsigned int conCounter=0; conCounter<constants_list.size(); conCounter++){
      delete constants_list[conCounter];
  }
  return total;
}

// This should eventually use a genome member function rather than an external.
// When we pack/unpack a population we also stuff its statistics.
int
SendPopulation(int toid, const GAPopulation& pop) {
	//cerr << "DEBUG:  started genome::SendPopulation\n";
  int status = 0;
  int psize = pop.size();
	int numbits = psize; /*This is for per-individual '# bits' ints*/
	int * intbuf = 0;
	int bufindex = 0;
	float * floatbuf = (float*)malloc(psize * sizeof(float));
	int sizes[2];
	if (0 == floatbuf) {
		cerr << "ERROR: malloc error: floatbuf in genome::SendPopulation\n";
		exit(1);
	}

	/*Determine how many bits are in each generation*/
	for (int i = 0; i < pop.size(); i++) {
		numbits += ((GEBinStrBlockCrossGenome&)(pop.individual(i))).length();
	}

	cerr << "DEBUG: numbits: " << numbits << "\n";

	intbuf = (int*)malloc(numbits * sizeof(int));
	if (0 == intbuf) {
		cerr << "ERROR:  malloc error: intbuf in genome::SendPopulation\n";
		exit(1);
	}
	
	for (int i = 0; i < pop.size(); i++) {
		bufindex = PackIndividualBits(pop.individual(i), intbuf, bufindex);
	}

	//cerr << "DEBUG: SendPopulation Bits:\n";
	//for (int i = 0; i < numbits; i++) {
		//cerr << "DEBUG: " <<  intbuf[i] << "\n";
	//}

	/*Send the scores as a seperate packet*/
	for (int i = 0; i < pop.size(); i++) {
		float score = ((GEBinStrBlockCrossGenome&)(pop.individual(i))).score();
		//printf("Storing score %f\n", score);
		floatbuf[i] = score;
	}

	sizes[0] = psize;
	sizes[1] = numbits;
	
	status = MPI_Send(sizes, 2, MPI_INT, toid, MSG_INCOMING_POPULATION, 
										MPI_COMM_WORLD);

	status = MPI_Send(intbuf, numbits, MPI_INT, toid, 
															MSG_INCOMING_POPULATION_BITS, MPI_COMM_WORLD);

	status = MPI_Send(floatbuf, pop.size(), MPI_FLOAT, toid, 
										MSG_INCOMING_POPULATION_SCORES, MPI_COMM_WORLD);


	free(intbuf);
	free(floatbuf);
	
	return status;
}


// This assumes that the original population contains at least one individual
// from which to grow.  If it does not, the data in the buffer will be ignored.
int
RecvPopulation(GAPopulation& pop, int tid) {
	//cerr << "DEBUG:  started genome::RecvPopulation\n";
  int status = 0;
  //int psize = 0;

	int sizes[2];
	MPI_Status Stat;
	int * intbuf;
	int bufindex = 0;
	float * floatbuf;

	status = MPI_Recv(&sizes, 2, MPI_INT, tid, MSG_INCOMING_POPULATION,
										MPI_COMM_WORLD, &Stat);

	intbuf = (int*)malloc(sizes[1] * sizeof(int));
	floatbuf = (float*)malloc(sizes[0] * sizeof(float));

	if ( (0 == intbuf) || (0 == floatbuf)) {
		printf("ERROR: memory error in genome.cpp::RecvPopulation\n");
		exit(1);
	}

	status = MPI_Recv(intbuf, sizes[1], MPI_INT, tid, 
															MSG_INCOMING_POPULATION_BITS,
															MPI_COMM_WORLD, &Stat);

	status = MPI_Recv(floatbuf, sizes[0], MPI_FLOAT, tid,
															MSG_INCOMING_POPULATION_SCORES,
															MPI_COMM_WORLD, &Stat);

	pop.size(sizes[0]);
	for (int i = 0; i < pop.size(); i++) {
		bufindex = UnpackIndividualBits(pop.individual(i), intbuf, bufindex);
		((GEBinStrBlockCrossGenome&)(pop.individual(i))).score(floatbuf[i]);
	}
	
	free(intbuf);
	free(floatbuf);
	return status;
}

// Send the specified number of individuals from the current population to
// the specified task.
int SendMigration(int * toids, GAGeneticAlgorithm& ga, int count, int numIds) {
  int status = 0;

	int numbits = count;  /*per-individual '# bits' ints*/
	int * intbuf = 0;
	int bufindex = 0;
	float * floatbuf = (float*)malloc(count * sizeof(float));
	int sizes[2];
	if (0 == floatbuf) {
		cerr << "ERROR:  malloc returned 0 for floatbuf in genome::SendMigration\n";
		exit(1);
	}

	/*Determine how many bits are in the individuals we're sending*/
	for (int i = 0; i < count; i++) {
		numbits += ((GEBinStrBlockCrossGenome&)(ga.population().best(i))).length();
		floatbuf[i] = ((GEBinStrBlockCrossGenome&)(ga.population().best(i))).score();
	}

	intbuf = (int*)malloc(numbits * sizeof(int));
	if (0 == intbuf) {
		cerr << "ERROR:  malloc returned 0 for intbuf in genome::SendMigration\n";
		exit(1);
	}

	/*Pack the bits into an array*/
	for (int i = 0; i < count; i++) {
		bufindex = PackIndividualBits(ga.population().best(i), intbuf, bufindex);
	}

	sizes[0] = count;
	sizes[1] = numbits;

	/*Send to each of them*/
	for (int idindex = 0; idindex < numIds; idindex++) {
		int toid = toids[idindex];


		status = MPI_Send(sizes, 2, MPI_INT, toid, MSG_INCOMING_MIGRATION,
											MPI_COMM_WORLD);
		
		status = MPI_Send(intbuf, numbits, MPI_INT, toid,
																MSG_INCOMING_MIGRATION_BITS,
																MPI_COMM_WORLD);
		
		status = MPI_Send(floatbuf, count, MPI_FLOAT, toid,
																MSG_INCOMING_MIGRATION_SCORES,
																MPI_COMM_WORLD);

	} /*end to each pop*/
		
	free(intbuf);
	free(floatbuf);

	return status;

}  /*end SendMigration*/


// Receive a bunch of individuals from a task.  To do this, we clone the GA's
// population, stuff the immigrants into the population, trash the worst 
// individuals to bring the population size back down to what it was, then
// stick the population back into the GA.
//   This implementation is really inefficient, but you get the idea of how
// to do this...
int RecvMigration(GAGeneticAlgorithm& ga, int tid) {
  int status = 0;
	int sizes[2];
	MPI_Status Stat;
	int bufindex = 0;
	int * intbuf = 0;
	float * floatbuf = 0;

	//cerr << "DEBUG:  starting genome::RecvMigration from " << tid << "\n";

  GAPopulation pop(ga.population());
  GAGenome *tmpind = ga.population().individual(0).clone();

	status = MPI_Recv(&sizes, 2, MPI_INT, tid, MSG_INCOMING_MIGRATION,
										MPI_COMM_WORLD, &Stat);

	intbuf = (int*)malloc(sizes[1] * sizeof(int));
	floatbuf = (float*)malloc(sizes[0] * sizeof(float));

	if ( (0 == intbuf) || (0 == floatbuf)) {
		printf("ERROR: memory error in genome.cpp::RecvMigration\n");
		exit(1);
	}

	status = MPI_Recv(intbuf, sizes[1], MPI_INT, tid,
															MSG_INCOMING_MIGRATION_BITS,
															MPI_COMM_WORLD, &Stat);

	status = MPI_Recv(floatbuf, sizes[0], MPI_FLOAT, tid,
															MSG_INCOMING_MIGRATION_SCORES,
															MPI_COMM_WORLD, &Stat);

	for (int i = 0; i < sizes[0]; i++) {
		bufindex = UnpackIndividualBits(*tmpind, intbuf, bufindex);
		((GEBinStrBlockCrossGenome&)(*tmpind)).score(floatbuf[i]);
		pop.add(*tmpind);		
	}

  for(int j=0; j<sizes[0]; j++)
    pop.destroy();		// default behavior is to destroy worst

  ga.population(pop);
  delete tmpind;

	free(intbuf);
	free(floatbuf);

  return status;
}

bool fullTree=true;

/* ----------------------------------------------------------------------------
  SensibleInitializer
---------------------------------------------------------------------------- */
// Sets trees to be 
void SensibleInitializer(GAGenome & c)
{
  GEBinStrBlockCrossGenome &child=DYN_CAST(GEBinStrBlockCrossGenome &, c);

  int maxDepth = gram.get_init_tree_depth();

  // create tree that initializes genome
  // tree will be full half the time and grown half the time                 
  // as described by koza
  fullTree = !fullTree;
  tree_node start;

  start.nodeElement = gram.get_start();
  start.ruleIndex = get_rule_index(start.nodeElement, maxDepth, fullTree);            

  tree<tree_node> tr(start);        
  tree<tree_node>::iterator top;
  top=tr.begin(); 

  add_rule_nodes(tr, top, maxDepth, fullTree);

  // after tree is constructed need to build the linear version
  // of the tree as the initial genome               
  vector<unsigned int> codons;
  tree<tree_node>::iterator currNode;
  int totalPossible = int(pow(float(2), int(child.CODONSIZE)));
  int maxNum, numRules, newCodonValue;
  for(currNode = tr.begin(); currNode != tr.end(); currNode++){
  // generate value within codoonsize that will give modulus desired
    numRules = gram.get_num_rules((*currNode).nodeElement);
    maxNum = totalPossible / numRules;
    newCodonValue = (*currNode).ruleIndex + GARandomInt(0, maxNum-1) * numRules;
    codons.push_back(newCodonValue);
  }
  // check to see if tree exceeds max size of genome
  // when tree exceeds number of codons, fill up as far as possible within the 
  // maximum size
  unsigned int lastCodon;
  if(tr.size() * child.CODONSIZE > child.maxX){
    lastCodon = child.maxX / child.CODONSIZE;
    child.resize(child.maxX);
  }
  else{
    // now resize genome to be the size needed
    child.resize(codons.size() * child.CODONSIZE);
    lastCodon = codons.size();
  }
  
  // now use masks to test bits in the codons vector and
  // set the bits to match
  // start at highest bit position
  vector<unsigned int> masks;
  for(int i=child.CODONSIZE-1; i>=0; i--){
    masks.push_back((unsigned int)pow(float(2), i));
  }
 
  // test each codon for each mask and set the
  // genes to match
  vector<unsigned int>::iterator maskIter;

  unsigned int geneCounter=0;
  // found and fixed problem on 3/8/05 smd
  // was not correctly setting genes
  for(unsigned int currCodon=0; currCodon < lastCodon; currCodon++){
    for(maskIter=masks.begin(); maskIter != masks.end(); maskIter++){
      child.gene(geneCounter++, *maskIter & codons[currCodon]);
    } 
  }
  
}


void add_rule_nodes(tree<tree_node> & tr, tree<tree_node>::iterator current_node,
  int depth, bool fullTree){
 
  tree<tree_node>::iterator newNodeIter;
 
  // check each nonterminal element in a randomly selected rule for the
  // current node and add those to the tree
  tree_node currNode = *current_node;
  int nodeIndex = get_rule_index(currNode.nodeElement, depth, fullTree);
  (*current_node).ruleIndex = nodeIndex;
  tree_node newNode;
 
  vector<Element *> elementRules = gram.get_rules(nodeIndex, currNode.nodeElement);
  vector<Element *>::iterator ruleIter;
  for(ruleIter=elementRules.begin(); ruleIter!=elementRules.end(); ruleIter++){
    if(!(*ruleIter)->get_isterminal()){
      newNode.nodeElement = *ruleIter;
      newNodeIter = tr.append_child(current_node, newNode);
      add_rule_nodes(tr, newNodeIter, depth-1, fullTree);
    }
  }
}


// returns the ruleIndex to use
int get_rule_index(Element * currentElement, int maxDepth, bool fullTree){
  int randIndex, numOpts;
  if(fullTree==true){
    numOpts = gram.get_num_recursive_opts(currentElement, maxDepth);
    if(numOpts > 0){
      randIndex = GARandomInt(0, numOpts-1);
      return gram.get_real_recursive_index(currentElement, maxDepth, randIndex);
    }
  }
  // below is for when no recursive options
  numOpts = gram.get_num_opts(currentElement, maxDepth);
  randIndex = GARandomInt(0, numOpts-1);
  return gram.get_real_index(currentElement, maxDepth, randIndex);
}

/*See PackIndividual - this one just does the bits.  Stores into given int
  array, starting at given position.  Returns next free position in 
  the buffer.*/
int PackIndividualBits(GAGenome& g, int * intbuf, int bufstart) {
	
  GEBinStrBlockCrossGenome& genome = (GEBinStrBlockCrossGenome&)g;

  int length = genome.length();
	intbuf[bufstart++] = length;


	int i = 0;
	for (i = 0; i < length; i++) {
		intbuf[bufstart + i] = genome.gene(i);
	}

	return bufstart + i;
  
}

/* See UnpackIndividual - this is the counterpart to PackIndividualBits
   unpacks an individual from intbuf, starting at position bufstart, 
   storing the information in the given genome.*/
int UnpackIndividualBits(GAGenome& g, int * intbuf, int bufstart) {
	
  GEBinStrBlockCrossGenome& genome = (GEBinStrBlockCrossGenome&)g;

	int* bits = 0;
	int length = intbuf[bufstart++]; /*bufstart now points to start of*/
	genome.length(length);           /* bits, not length*/
	bits = new int [length];

	int i = 0;
	for (i = 0; i < length; i++) {
		bits[i] = intbuf[bufstart + i];
	}

	genome = bits;

	delete [] bits;

	return bufstart + i;

} /*end UnpackIndividualBits*/
